#!/usr/bin/env python
# coding=utf-8

__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Standard library imports
import os
import sys
import time
import logging
import ast
import json
import re

# Networking and URL handling imports
from urllib.parse import urlencode
import urllib3

# Disable insecure request warnings for urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# appebd lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# Import trackme libs logicalgroup
from trackme_libs_logicalgroup import logical_group_update_green_red_members

# logging:
# To avoid overriding logging destination of callers, the libs will not set on purpose any logging definition
# and rely on callers themselves


def parse_filters(query_parameters):
    filters = []
    for key, value in query_parameters.items():
        if "filter[" in key:
            parts = key.split("[")
            index = int(parts[1].split("]")[0])
            prop = parts[2].split("]")[0]

            # Initialize filter dict if it doesn't exist
            while len(filters) <= index:
                filters.append({})

            if "value" in key and len(parts) > 3:
                # Handle list values
                if "value" not in filters[index]:
                    filters[index]["value"] = []
                # Convert list index (e.g., filter[0][value][0]) to int and insert value
                list_index = int(parts[3].split("]")[0])
                # Ensure the list is long enough to hold this index
                while len(filters[index]["value"]) <= list_index:
                    filters[index]["value"].append(None)
                filters[index]["value"][list_index] = (
                    value.lower() if isinstance(value, str) else value
                )
            else:
                # Handle non-list values or the property itself
                filters[index][prop] = (
                    value.lower() if isinstance(value, str) else value
                )
    return filters


def record_matches_filter(record, filter):
    field = filter.get("field")
    filter_type = filter.get("type")
    value = filter.get("value")

    # Immediately return True if the filter value is empty
    if value == "":
        return True

    # Try to interpret the value as a JSON list if it looks like one
    if isinstance(value, str) and value.startswith("[") and value.endswith("]"):
        try:
            value = json.loads(value)
        except json.JSONDecodeError:
            pass  # If decoding fails, proceed with the original value string

    # Prepare the record value for comparison
    record_value = record.get(field, "")
    if (
        isinstance(record_value, str)
        and "|" in record_value
        and field not in ["alias", "object"]
    ):
        # Treat as a pseudo list if record_value contains pipes
        record_value = [item.strip().lower() for item in record_value.split("|")]
    elif isinstance(record_value, str):
        record_value = record_value.strip().lower()

    if isinstance(value, str):
        value = value.strip().lower()

    # Handling for different filter types when record_value is a list
    if isinstance(record_value, list):
        if filter_type == "like":
            if isinstance(value, list):
                return any(v.lower() in item for item in record_value for v in value)
            else:
                return any(value in item for item in record_value)
        elif filter_type == "=":
            if isinstance(value, list):
                return any(item == v.lower() for item in record_value for v in value)
            else:
                return value in record_value
        # we can accept != as a filter
        elif filter_type == "!=":
            if isinstance(value, list):
                return any(item != v.lower() for item in record_value for v in value)
            else:
                return value not in record_value
        elif filter_type in ("<", "<=", ">", ">="):
            # Numerical comparisons for lists are more complex and context-dependent
            # You might want to reconsider how these should behave with list record_values
            return False
        elif filter_type == "in":
            if isinstance(value, list):
                return any(item in [v.lower() for v in value] for item in record_value)
            else:
                return value in record_value
        elif filter_type == "starts":
            return any(item.startswith(value) for item in record_value)
        elif filter_type == "ends":
            return any(item.endswith(value) for item in record_value)
        elif filter_type == "regex":
            return any(re.search(value, item) is not None for item in record_value)
    else:
        # Handling for different filter types when record_value is a string
        if filter_type == "like":
            return value in record_value
        elif filter_type == "=":
            return record_value == value
        # numerical comparison (except for != which is handled as a string comparison)
        elif filter_type in ("<", "<=", ">", ">=", "!="):
            # Handle numerical comparisons including "!="
            if filter_type == "!=":
                try:
                    # Attempt numerical comparison first
                    is_not_equal = float(record_value) != float(value)
                except ValueError:
                    # Fallback to string comparison
                    is_not_equal = record_value != value
                return is_not_equal
            else:
                try:
                    record_value = float(record_value)
                    value = float(value)
                except ValueError:
                    return False  # Skip filter if conversion fails
                # Directly evaluate the comparison expression
                return eval(f"record_value {filter_type} value")
        elif filter_type == "in":
            if isinstance(value, list):
                return record_value in [v.lower() for v in value]
            else:
                return record_value == value
        elif filter_type == "starts":
            return record_value.startswith(value)
        elif filter_type == "ends":
            return record_value.endswith(value)
        elif filter_type == "regex":
            return re.search(value, record_value) is not None

    return False


def pre_filter_records(data_records, query_parameters):
    """
    Pre-filters records based on a subset of filter fields: 'alias', 'object', 'monitored_state'.
    If other fields are present in the filters, returns all records without filtering.
    """
    filters = parse_filters(query_parameters)
    prefilter_fields = {"alias", "object", "monitored_state"}

    # Check if any filter exists outside the pre-defined fields for pre-filtering
    if any(f.get("field") not in prefilter_fields for f in filters):
        # If there are filters on fields outside the pre-filter scope, return all records
        return data_records

    # Proceed with pre-filtering if all filters fall within the pre-filter scope
    prefiltered_records = [
        record
        for record in data_records
        if all(record_matches_filter(record, f) for f in filters)
    ]

    return prefiltered_records


def filter_records(data_records, query_parameters):
    """
    Filters data records based on structured filters parsed from query parameters.
    """
    filters = parse_filters(query_parameters)

    if len(filters) > 0:
        logging.debug(f'filters="{filters}"')

    # Apply filters to records, requiring all conditions to be met ('AND' logic)
    filtered_records = [
        record
        for record in data_records
        if all(record_matches_filter(record, f) for f in filters)
    ]

    return filtered_records


def convert_seconds_to_duration(seconds):
    """
    Define the function convert_seconds_to_duration
    behaviour: converts seconds to duration, duration is a string from as [D+]HH:MM:SS
    The first segment represents the number of days, the second the number of hours, third the number of minutes, and the fourth the number of seconds.
    """

    try:
        original_seconds = int(seconds)
    except ValueError:
        return 0

    # Check if the original seconds were negative
    is_negative = original_seconds < 0
    seconds = abs(original_seconds)

    # Calculate days, hours, minutes, and seconds
    days = seconds // (24 * 3600)
    seconds = seconds % (24 * 3600)
    hours = seconds // 3600
    seconds %= 3600
    minutes = seconds // 60
    seconds %= 60

    # Format the duration string
    if days > 0:
        duration = f"{days}+{hours:02d}:{minutes:02d}:{seconds:02d}"
    else:
        duration = f"{hours:02d}:{minutes:02d}:{seconds:02d}"

    # Add "-" if the original seconds were negative
    if is_negative:
        duration = "-" + duration

    return duration


def convert_epoch_to_datetime(epoch):
    """
    Define the function convert_epoch_to_datetime
    """

    # convert epoch to float
    try:
        epoch = float(epoch)
        # convert epoch to datetime
        datetime = time.strftime("%d %b %Y %H:%M", time.localtime(epoch))
        return datetime
    except Exception as e:
        epoch = 0
        return epoch


def get_monitoring_wdays(monitoring_wdays, monitoring_hours_ranges):
    """
    Create a new function called get_monitoring_wdays which returns True or False
    arguments: monitoring_wdays, monitoring_hours_ranges
    """

    try:

        # get current wday
        current_wday_no = int(time.strftime("%w"))

        # get current hour range
        current_hour_range = int(time.strftime("%H"))

        # get manual wdays
        manual_wdays_pattern = monitoring_wdays.replace("manual:", "").replace(",", "|")
        manual_wdays_regex = re.compile(f"^{manual_wdays_pattern}$")

        # get manual hours ranges
        manual_hours_pattern = monitoring_hours_ranges.replace("manual:", "").replace(
            ",", "|"
        )
        manual_hours_regex = re.compile(f"^({manual_hours_pattern})$")

        # apply wdays rules
        if monitoring_wdays == "auto:all_days" or monitoring_wdays == "manual:all_days":
            isUnderMonitoringDays = True
        elif (
            monitoring_wdays == "auto:monday-to-friday"
            or monitoring_wdays == "manual:monday-to-friday"
        ) and current_wday_no in [
            1,
            2,
            3,
            4,
            5,
        ]:
            isUnderMonitoringDays = True
        elif (
            monitoring_wdays == "auto:monday-to-saturday"
            or monitoring_wdays == "manual:monday-to-saturday"
        ) and current_wday_no in [
            1,
            2,
            3,
            4,
            5,
            6,
        ]:
            isUnderMonitoringDays = True
        elif monitoring_wdays.startswith("manual:") and manual_wdays_regex.match(
            str(current_wday_no)
        ):
            isUnderMonitoringDays = True
        else:
            isUnderMonitoringDays = False

        # apply hours ranges rules
        if (
            monitoring_hours_ranges == "auto:all_ranges"
            or monitoring_hours_ranges == "manual:all_ranges"
        ):
            isUnderMonitoringHours = True
        elif (
            monitoring_hours_ranges == "auto:08h-to-20h"
            or monitoring_hours_ranges == "manual:08h-to-20h"
        ) and current_hour_range in range(8, 19):
            isUnderMonitoringHours = True
        elif monitoring_hours_ranges.startswith("manual:") and manual_hours_regex.match(
            str(current_hour_range)
        ):
            isUnderMonitoringHours = True
        else:
            isUnderMonitoringHours = False

        # generate a field isUnderMonitoringMsg (string) as a summary of the monitoring status
        if isUnderMonitoringDays is True and isUnderMonitoringHours is True:
            isUnderMonitoringMsg = f'''"This entity is currently under monitoring for days {monitoring_wdays} and hours {monitoring_hours_ranges}"'''

        elif isUnderMonitoringDays is False and isUnderMonitoringHours is True:
            isUnderMonitoringMsg = f'''This entity is not currently under monitoring days conditions, days="{monitoring_wdays}", hours="{monitoring_hours_ranges}"'''

        elif isUnderMonitoringDays is True and isUnderMonitoringHours is False:
            isUnderMonitoringMsg = f'''This entity is not currently under monitoring hours conditions, days="{monitoring_wdays}", hours="{monitoring_hours_ranges}"'''

        else:
            isUnderMonitoringMsg = f'''This entity is not currently under monitoring days and hours conditions, days="{monitoring_wdays}", hours="{monitoring_hours_ranges}"'''

        # return
        return isUnderMonitoringDays, isUnderMonitoringHours, isUnderMonitoringMsg

    except Exception as e:
        logging.error(f"get_monitoring_wdays function has failed, exception={str(e)}")
        return (
            True,
            True,
            f"get_monitoring_wdays function has failed, exception={str(e)}",
        )


def get_outliers_status(isOutlier, OutliersDisabled, tenant_outliers_set_state):
    """
    Create a function called get_outliers_status:
    - arguments: isOutlier, OutliersDisabled, tenant_outliers_set_state
    - tenant_outliers_set_state is 0 or 1, 0 means it is disabled at tenant level, 1 means it is enabled at tenant level
    - behaviour: alter isOutlier, 0=not outlier, 1=outlier, 2=outlier but disabled at tenant level
    """

    if OutliersDisabled == 1:
        isOutlier = 0

    else:

        if tenant_outliers_set_state == 1:
            if isOutlier == 1:
                isOutlier = 1
            else:
                isOutlier = 0
        else:
            if isOutlier == 1:
                isOutlier = 2
            else:
                isOutlier = 0

    return isOutlier


def get_data_sampling_status(isAnomaly, tenant_data_sampling_set_state):
    """
    Create a function called get_data_sampling_status:
    - arguments: isAnomaly, tenant_data_sampling_set_state
    - tenant_data_sampling_set_state is 0 or 1, 0 means it is disabled at tenant level, 1 means it is enabled at tenant level
    - behaviour: alter isAnomaly, 0=not anomaly, 1=anomaly, 2=anomaly but disabled at tenant level
    """

    if tenant_data_sampling_set_state == 1:
        if isAnomaly == 1:
            isAnomaly = 1
        else:
            isAnomaly = 0
    else:
        if isAnomaly == 1:
            isAnomaly = 2
        else:
            isAnomaly = 0

    return isAnomaly


def get_future_status(
    future_tolerance,
    system_future_tolerance,
    data_last_lag_seen,
    data_last_ingestion_lag_seen,
    data_last_time_seen,
    data_last_ingest,
):
    """
    Create a function called get_future_status:
    - arguments: future_tolerance (expressed in seconds), system_future_tolerance (expressed in seconds), data_last_lag_seen (expressed in seconds)
    - behaviour: returns a boolean, True if data_last_lag_seen is lower than future_tolerance, False otherwise. If future_tolerance is 0, rely on system_future_tolerance
    """

    isFuture = False
    isFutureMsg = ""
    if future_tolerance == 0:
        future_tolerance = system_future_tolerance

    # convert all to int
    try:
        future_tolerance = int(round(float(future_tolerance), 0))
        data_last_lag_seen = int(round(float(data_last_lag_seen), 0))
        data_last_ingestion_lag_seen = int(
            round(float(data_last_ingestion_lag_seen), 0)
        )
    except:
        pass

    logging.debug(
        f"data_last_lag_seen={data_last_lag_seen}, system_future_tolerance={system_future_tolerance}, future_tolerance={future_tolerance}"
    )

    if float(data_last_lag_seen) < float(future_tolerance) or float(
        data_last_ingestion_lag_seen
    ) < float(future_tolerance):
        isFuture = True

        # convert data_last_lag_seen to duration
        data_last_lag_seen_duration = convert_seconds_to_duration(data_last_lag_seen)

        # convert data_last_ingestion_lag_seen to duration
        data_last_ingestion_lag_seen_duration = convert_seconds_to_duration(
            data_last_ingestion_lag_seen
        )

        # convert data_last_time_seen to %c
        data_last_time_seen_datetime = convert_epoch_to_datetime(data_last_time_seen)

        # convert data_last_ingest to %c
        data_last_ingest_datetime = convert_epoch_to_datetime(data_last_ingest)

        isFutureMsg = f"""detected data indexed in the future which is most likely due to timestamping misconfiguration, timezone or time synchronization issue. Event delay is {data_last_lag_seen} seconds (duration: {data_last_lag_seen_duration}), Event latency is {data_last_ingestion_lag_seen} seconds (duration: {data_last_ingestion_lag_seen_duration}), this is beyond current tolerance threshold of {future_tolerance} seconds, latest event available (_time) for this entity: {data_last_time_seen_datetime}, latest event ingested for this entity: {data_last_ingest_datetime}. Review and fix the root cause, or adapt the future tolerance at the system level or for this entity especially."""

    else:
        isFuture = False

    return isFuture, isFutureMsg, future_tolerance


def get_future_metrics_status(
    system_future_tolerance,
    metric_last_time_seen,
):
    """
    Create a function called get_future_metrics_status:
    - arguments: future_tolerance (expressed in seconds), system_future_tolerance (expressed in seconds), data_last_lag_seen (expressed in seconds)
    - behaviour: returns a boolean, True if metric_last_time_seen is lower than system_future_tolerance, False otherwise.
    """

    isFuture = False
    isFutureMsg = ""

    logging.debug(
        f"metric_last_time_seen={metric_last_time_seen}, system_future_tolerance={system_future_tolerance}"
    )

    if float(metric_last_time_seen) < float(system_future_tolerance):
        isFuture = True
        # convert data_last_lag_seen to duration
        metric_last_time_seen_duration = convert_seconds_to_duration(
            metric_last_time_seen
        )
        # convert data_last_time_seen to %c
        metric_last_time_seen_datetime = convert_epoch_to_datetime(
            metric_last_time_seen
        )

        isFutureMsg = f"""detected metrics indexed in the future which is most likely due to timestamping misconfiguration, timezone or time synchronization issue. Metric delay is {metric_last_time_seen} seconds (duration: {metric_last_time_seen_duration}) which is beyond tolerance threshold of {system_future_tolerance}, latest event available (_time) for this entity: {metric_last_time_seen_datetime}"""
    else:
        isFuture = False

    return isFuture, isFutureMsg


def get_is_under_dcount_host(min_dcount_host, min_dcount_threshold, min_dcount_field):
    """
    Create a function call get_is_under_dcount_host:
    - arguments: min_dcount_host, min_dcount_threshold
    - returns: isUnderDcountHost (boolean), isUnderDcountHostMsg (string)
    - behaviour: returns a boolean, True if min_dcount_threshold is a numerical and is lower than min_dcount_host, False otherwise
    """

    if isinstance(min_dcount_host, float) and min_dcount_threshold < min_dcount_host:
        isUnderDcountHost = True
        isUnderDcountHostMsg = f"""Monitoring conditions are not met due to low number of hosts. Number of hosts is {min_dcount_threshold} based on the metric {min_dcount_field} which is lower than the minimum required number of hosts of {min_dcount_host}"""
    else:
        isUnderDcountHost = False
        isUnderDcountHostMsg = ""

    return isUnderDcountHost, isUnderDcountHostMsg


def get_logical_groups_collection_records(collection):
    """
    Queries and processes records from a collection based on specific criteria.

    :param collection: The collection object to query.
    :return: Tuple containing collection records and a dictionary of records.
    """

    collection_records = []
    collection_records_dict = {}
    count_to_process_list = []
    collection_members_list = []
    collection_members_dict = {}

    end = False
    skip_tracker = 0
    while not end:
        process_collection_records = collection.data.query(skip=skip_tracker)
        if process_collection_records:
            for item in process_collection_records:
                collection_records.append(item)
                collection_records_dict[item.get("_key")] = {
                    "object_group_name": item.get("object_group_name"),
                    "object_group_mtime": item.get("object_group_mtime"),
                    "object_group_members": item.get("object_group_members", []),
                    "object_group_members_green": item.get(
                        "object_group_members_green", []
                    ),
                    "object_group_members_red": item.get(
                        "object_group_members_red", []
                    ),
                    "object_group_min_green_percent": item.get(
                        "object_group_min_green_percent", 0
                    ),
                }

                try:

                    logicalgroup_members = item.get("object_group_members", [])
                    # add members in collection_members_list, also create a dict per member
                    for member in logicalgroup_members:
                        if member not in collection_members_list:
                            collection_members_list.append(member)
                            collection_members_dict[member] = {
                                "object_group_key": item.get("_key"),
                                "object_group_name": item.get("object_group_name"),
                                "object_group_members": item.get(
                                    "object_group_members", []
                                ),
                                "object_group_members_green": item.get(
                                    "object_group_members_green", []
                                ),
                                "object_group_members_red": item.get(
                                    "object_group_members_red", []
                                ),
                                "object_group_min_green_percent": item.get(
                                    "object_group_min_green_percent", 0
                                ),
                            }

                except Exception as e:
                    logging.error(
                        f"function get_logical_groups_collection_records, error while processing logical group members, exception={str(e)}"
                    )

                count_to_process_list.append(item.get("_key"))
            skip_tracker += 5000
        else:
            end = True

    return (
        collection_records,
        collection_records_dict,
        collection_members_list,
        collection_members_dict,
        count_to_process_list,
    )


def get_and_manage_logical_group_status(
    splunkd_uri,
    session_key,
    tenant_id,
    object_name,
    object_state,
    object_group_key,
    object_logical_group_dict,
):
    """
    Create a function called get_and_manage_logical_group_status:
    - arguments: object_name, object_state, object_group_key, object_logical_group_dict
    - returns: isUnderLogicalGroup (boolean), LogicalGroupStateInAlert (boolean), LogicalGroupMsg (string)
    - behaviour:
        isUnderLogicalGroup: True if object_group_key is not empty and object_group_members_count is higher than 1
        LogicalGroupStateInAlert: True if object_group_green_percent is lower than object_group_min_green_percent
        LogicalGroupMsg: string containing the status of the logical group
    """

    # object_group_members_count
    object_group_members_count = 0

    # get logical group name
    object_group_name = object_logical_group_dict.get("object_group_name", "")

    try:

        # enter if the group is not empty
        if object_group_name != "":
            object_group_min_green_percent = object_logical_group_dict.get(
                "object_group_min_green_percent", 0
            )

            object_group_members = object_logical_group_dict.get(
                "object_group_members", []
            )
            try:
                object_group_members_count = len(object_group_members)
            except:
                object_group_members_count = 0

            # if not a list and is a string, convert to list
            if isinstance(object_group_members, str):
                object_group_members = [object_group_members]

            object_group_members_green = object_logical_group_dict.get(
                "object_group_members_green", []
            )
            # if not a list and is a string, convert to list
            if isinstance(object_group_members_green, str):
                object_group_members_green = [object_group_members_green]

            object_group_members_red = object_logical_group_dict.get(
                "object_group_members_red", []
            )
            # if not a list and is a string, convert to list
            if isinstance(object_group_members_red, str):
                object_group_members_red = [object_group_members_red]

            # if object_state is green, object_name must be in object_group_members_green but not in object_group_members_red
            # if object_state is red or blue, object_name must be in object_group_members_red but not in object_group_members_green
            # any change required to object_group_members_green and object_group_members_red implies an update to the KVstore record is required, set the boolean uppdate_kvstore_record to True if required

            update_kvstore_record = False

            if object_state == "green":
                if object_name not in object_group_members_green:
                    object_group_members_green.append(object_name)
                    update_kvstore_record = True
                if object_name in object_group_members_red:
                    object_group_members_red.remove(object_name)
                    update_kvstore_record = True
            else:
                if object_name not in object_group_members_red:
                    object_group_members_red.append(object_name)
                    update_kvstore_record = True
                if object_name in object_group_members_green:
                    object_group_members_green.remove(object_name)
                    update_kvstore_record = True

            # if update_kvstore_record is True, call the API endpoint accordingly
            if update_kvstore_record:
                # proceed
                try:
                    response = logical_group_update_green_red_members(
                        splunkd_uri,
                        session_key,
                        tenant_id,
                        object_name,
                        object_group_key,
                        object_group_members_green,
                        object_group_members_red,
                    )
                    logging.info(
                        f'tenant="{tenant_id}", object="{object_name}", logical group green/red members update API was successfull, response="{response}"'
                    )

                except Exception as e:
                    logging.error(
                        f'tenant="{tenant_id}", object="{object_name}", logical group green/red members update API call has failed, exception="{str(e)}"'
                    )

            # ensure object_group_min_green_percent is float
            try:
                object_group_min_green_percent = float(object_group_min_green_percent)
            except:
                object_group_min_green_percent = 0

            # calculate object_group_green_percent, if logical group is empty, then object_group_green_percent is 100
            try:
                if object_group_members_count > 0:
                    object_group_green_percent = (
                        len(object_group_members_green) / object_group_members_count
                    ) * 100
                else:
                    object_group_green_percent = 100
            except:
                object_group_green_percent = 0

        # define status and return

        if object_group_key != "" and object_group_members_count > 1:
            isUnderLogicalGroup = True
            if object_group_green_percent < object_group_min_green_percent:
                LogicalGroupStateInAlert = True
                LogicalGroupMsg = f"""Logical Group {object_group_name} with key="{object_group_key}" is in alert state. The current green percentage of the group is {round(object_group_green_percent, 2)}% which is lower than the minimum green percentage of {round(object_group_min_green_percent, 2)}%, object_group_members_count={object_group_members_count}, object_group_members_red={object_group_members_red}"""

            else:
                LogicalGroupStateInAlert = False
                LogicalGroupMsg = f"""Logical Group {object_group_name} with key="{object_group_key}" is in normal state. The current green percentage of the group is {round(object_group_green_percent, 2)}% which is higher or equal to the minimal green percentage of {round(object_group_min_green_percent, 2)}%, object_group_members_count={object_group_members_count}, object_group_members_red={object_group_members_red}"""

        else:
            isUnderLogicalGroup = False
            LogicalGroupStateInAlert = False
            LogicalGroupMsg = ""

        return isUnderLogicalGroup, LogicalGroupStateInAlert, LogicalGroupMsg

    except Exception as e:
        logging.error(
            f'function get_and_manage_logical_group_status has failed, exception="{str(e)}", object_name="{object_name}", object_group_key="{object_group_key}", object_logical_group_dict="{object_logical_group_dict}"'
        )
        return (
            False,
            False,
            f'function get_and_manage_logical_group_status has failed, exception="{str(e)}", object_name="{object_name}", object_group_key="{object_group_key}"',
        )


def get_dsm_latency_status(
    data_last_ingestion_lag_seen,
    data_max_lag_allowed,
    data_lag_alert_kpis,
    data_last_ingest,
    data_last_time_seen,
):
    """
    Create a function called get_dsm_latency_status:
    - arguments: data_last_ingestion_lag_seen, data_max_lag_allowed, data_lag_alert_kpis, data_last_ingest, data_last_time_seen
    - returns: isUnderLatencyAlert (boolean), isUnderLatencyMessage (string)
    - behaviour:
        isUnderLatencyAlert: If data_lag_alert_kpis is "all_kpis" or "lag_ingestion_kpi" and if data_last_ingestion_lag_seen is higher than data_max_lag_allowed, then isUnderLatencyAlert is True
        isUnderLatencyMessage: If isUnderLatencyAlert is True:
        "Monitoring conditions are not met due to latency issues. Ingestion latency is $data_last_ingestion_lag_seen$ seconds (duration: <conversion of data_last_ingestion_lag_seen to duration), which is higher than the maximum allowed latency of $data_max_lag_allowed$ seconds (duration: <conversion of data_max_lag_allowed to duration>), latest event available for this entity: <convertion of epoch data_last_time_seen to %c>, latest event ingested for this entity: <convertion of epoch data_last_ingest to %c>"
        isUnderLatencyMessage: If isUnderLatencyAlert is False:
        "Monitoring conditions are not met for ingest latency are met. Ingestion latency is $data_last_ingestion_lag_seen$ seconds (duration: <conversion of data_last_ingestion_lag_seen to duration), which is lower than the maximum allowed latency of $data_max_lag_allowed$ seconds (duration: <conversion of data_max_lag_allowed to duration>), latest event available for this entity: <convertion of epoch data_last_time_seen to %c>, latest event ingested for this entity: <convertion of epoch data_last_ingest to %c>"
    """

    # convert data_last_ingestion_lag_seen to float
    try:
        data_last_ingestion_lag_seen = float(data_last_ingestion_lag_seen)
    except:
        data_last_ingestion_lag_seen = 0

    # convert data_max_lag_allowed to float
    try:
        data_max_lag_allowed = float(data_max_lag_allowed)
    except:
        data_max_lag_allowed = 0

    # convert data_last_ingest to float
    try:
        data_last_ingest = float(data_last_ingest)
    except:
        data_last_ingest = 0

    # convert data_last_time_seen to float
    try:
        data_last_time_seen = float(data_last_time_seen)
    except:
        data_last_time_seen = 0

    # convert data_last_ingestion_lag_seen to duration
    data_last_ingestion_lag_seen_duration = convert_seconds_to_duration(
        data_last_ingestion_lag_seen
    )

    # convert data_max_lag_allowed to duration
    data_max_lag_allowed_duration = convert_seconds_to_duration(data_max_lag_allowed)

    # convert data_last_ingest to %c
    data_last_ingest_datetime = convert_epoch_to_datetime(data_last_ingest)

    # convert data_last_time_seen to %c
    data_last_time_seen_datetime = convert_epoch_to_datetime(data_last_time_seen)

    # define isUnderLatencyAlert
    if data_lag_alert_kpis in ["all_kpis", "lag_ingestion_kpi"]:
        if float(data_last_ingestion_lag_seen) > float(data_max_lag_allowed):
            isUnderLatencyAlert = True
        else:
            isUnderLatencyAlert = False
    else:
        isUnderLatencyAlert = False

    # define isUnderLatencyMessage
    if isUnderLatencyAlert:
        isUnderLatencyMessage = f"""Monitoring conditions are not met due to latency issues. Ingestion latency is approximately {round(float(data_last_ingestion_lag_seen), 3)} seconds (duration: {data_last_ingestion_lag_seen_duration}), which is higher than the maximum allowed latency of {round(float(data_max_lag_allowed), 0)} seconds (duration: {data_max_lag_allowed_duration}), latest event available (_time) for this entity: {data_last_time_seen_datetime}, latest event indexed (_indextime) for this entity: {data_last_ingest_datetime}"""
    else:
        isUnderLatencyMessage = f"""monitoring conditions for ingest latency are met. Ingestion latency is approximately {round(float(data_last_ingestion_lag_seen), 3)} seconds (duration: {data_last_ingestion_lag_seen_duration}), which is lower than the maximum allowed latency of {round(float(data_max_lag_allowed), 0)} seconds (duration: {data_max_lag_allowed_duration}), latest event indexed (_indextime) for this entity: {data_last_ingest_datetime}"""

    # return
    return isUnderLatencyAlert, isUnderLatencyMessage


def get_dsm_delay_status(
    data_last_lag_seen,
    data_max_delay_allowed,
    data_lag_alert_kpis,
    data_last_ingest,
    data_last_time_seen,
):
    """
    Create a function called get_dsm_delay_status:
    - arguments: data_last_lag_seen, data_max_delay_allowed, data_lag_alert_kpis, data_last_ingest, data_last_time_seen
    - returns: isUnderDelayAlert (boolean), isUnderDelayMessage (string)
    - behaviour:
        isUnderDelayAlert: If data_lag_alert_kpis is "all_kpis" or "lag_event_kpi" and if data_last_lag_seen is higher than data_max_delay_allowed, then isUnderDelayAlert is True
        isUnderDelayMessage: If isUnderDelayAlert is True:
            "Monitoring conditions are not met due to delay issues. Event delay is $data_last_lag_seen$ seconds (duration: <conversion of data_last_lag_seen to duration), which is higher than the maximum allowed delay of $data_max_delay_allowed$ seconds (duration: <conversion of data_max_delay_allowed to duration>), latest event available (_time) for this entity: <convertion of epoch data_last_time_seen to %c>, latest event ingested for this entity: <convertion of epoch data_last_ingest to %c>"
        isUnderDelayMessage: If isUnderDelayAlert is False:
            "monitoring conditions for event delay are met. Event delay is $data_last_lag_seen$ seconds (duration: <conversion of data_last_lag_seen to duration), which is lower than the maximum allowed delay of $data_max_delay_allowed$ seconds (duration: <conversion of data_max_delay_allowed to duration>), latest event available for this entity: <convertion of epoch data_last_time_seen to %c>, latest event ingested for this entity: <convertion of epoch data_last_ingest to %c>"
    """

    # convert data_last_lag_seen to float
    try:
        data_last_lag_seen = float(data_last_lag_seen)
    except:
        data_last_lag_seen = 0

    # convert data_max_delay_allowed to float
    try:
        data_max_delay_allowed = float(data_max_delay_allowed)
    except:
        data_max_delay_allowed = 0

    # convert data_last_ingest to float
    try:
        data_last_ingest = float(data_last_ingest)
    except:
        data_last_ingest = 0

    # convert data_last_time_seen to float
    try:
        data_last_time_seen = float(data_last_time_seen)
    except:
        data_last_time_seen = 0

    # convert data_last_lag_seen to duration
    data_last_lag_seen_duration = convert_seconds_to_duration(data_last_lag_seen)

    # convert data_max_delay_allowed to duration
    data_max_delay_allowed_duration = convert_seconds_to_duration(
        data_max_delay_allowed
    )

    # convert data_last_ingest to %c
    data_last_ingest_datetime = convert_epoch_to_datetime(data_last_ingest)

    # convert data_last_time_seen to %c
    data_last_time_seen_datetime = convert_epoch_to_datetime(data_last_time_seen)

    # define isUnderDelayAlert
    if data_lag_alert_kpis in ["all_kpis", "lag_event_kpi"]:
        if float(data_last_lag_seen) > float(data_max_delay_allowed):
            isUnderDelayAlert = True
        else:
            isUnderDelayAlert = False
    else:
        isUnderDelayAlert = False

    # define isUnderDelayMessage
    if isUnderDelayAlert:
        isUnderDelayMessage = f"""Monitoring conditions are not met due to delay issues. Event delay is {round(float(data_last_lag_seen), 3)} seconds (duration: {data_last_lag_seen_duration}), which is higher than the maximum allowed delay of {int(round(float(data_max_delay_allowed), 0))} seconds (duration: {data_max_delay_allowed_duration}), latest event available (_time) for this entity: {data_last_time_seen_datetime}, latest event ingested (_indextime) for this entity: {data_last_ingest_datetime}"""
    else:
        isUnderDelayMessage = f"""monitoring conditions for event delay are met. Event delay is {round(float(data_last_lag_seen), 3)} seconds (duration: {data_last_lag_seen_duration}), which is lower than the maximum allowed delay of {int(round(float(data_max_delay_allowed), 0))} seconds (duration: {data_max_delay_allowed_duration}), latest event available (_time) for this entity: {data_last_time_seen_datetime}"""

    # return
    return isUnderDelayAlert, isUnderDelayMessage


def set_dsm_status(
    splunkd_uri,
    session_key,
    tenant_id,
    record,
    isOutlier,
    isAnomaly,
    isFuture,
    isFutureMsg,
    isUnderMonitoringDays,
    isUnderMonitoringHours,
    isUnderMonitoringMsg,
    isUnderDcountHost,
    isUnderDcountHostMsg,
    object_logical_group_dict,
    isUnderLatencyAlert,
    isUnderLatencyMessage,
    isUnderDelayAlert,
    isUnderDelayMessage,
):
    """
    Create a function called set_dsm_status:
    - arguments: record, isOutlier, isAnomaly, isFuture, isUnderMonitoringDays, isUnderMonitoringHours, isUnderDcountHost, isUnderLogicalGroup, LogicalGroupStateInAlert, isUnderLatencyAlert, isUnderLatencyMessage, isUnderDelayAlert, isUnderDelayMessage
    - returns:
        object_state (string): blue, orange, green, red
        anomaly_reason (list): list of short code reasons why the object is in anomaly
        status_message (list): list of long description reasons why the object is in anomaly
    - behaviour:
        object_state:
            green if:
                isOutlier is 1
                isAnomaly is 1
                isFuture is False
                isUnderMonitoringDays is True
                isUnderMonitoringHours is True
                isUnderDcountHost is False
                if isUnderLogicalGroup is True, then LogicalGroupStateInAlert must be False
                isUnderLatencyAlert is False
                isUnderDelayAlert is False
            blue if:
                Any of the condition above is not met, but isUnderLogicalGroup is True and LogicalGroupStateInAlert is True
            orange if:
                All green conditions are met except for isFuture which would be True
            red if:
                Any of the green conditions are not met, and blue conditions and orange conditions are not met
        anomaly_reason:
            if object_state is green, anomnaly_reason is None
            Otherwise, anomaly_reason is a list containing the reasons why the object is in anomaly
    """

    # init status_message and anomaly_reason
    status_message = []
    anomaly_reason = []

    # init status_message_json
    status_message_json = {}

    # define object_state
    if (
        (isOutlier == 0 or isOutlier == 2)
        and (isAnomaly == 0 or isAnomaly == 2)
        and isUnderDcountHost is False
        and isUnderLatencyAlert is False
        and isUnderDelayAlert is False
    ):
        object_state = "green"
    else:
        object_state = "red"

    #
    # Logical group management
    #

    (
        isUnderLogicalGroup,
        LogicalGroupStateInAlert,
        LogicalGroupMsg,
    ) = get_and_manage_logical_group_status(
        splunkd_uri,
        session_key,
        tenant_id,
        record.get("object"),
        object_state,
        record.get("object_group_key"),
        object_logical_group_dict,
    )

    # log debug
    logging.debug(
        f'function get_and_manage_logical_group_status: object="{record.get("object")}", object_state="{object_state}", object_group_key="{record.get("object_group_key")}", isUnderLogicalGroup="{isUnderLogicalGroup}", LogicalGroupStateInAlert="{LogicalGroupStateInAlert}", LogicalGroupMsg="{LogicalGroupMsg}"'
    )

    # if object_state is red but isUnderLogicalGroup is True and LogicalGroupStateInAlert is False, then object_state is blue
    if object_state == "red" and isUnderLogicalGroup is True:
        if LogicalGroupStateInAlert is False:
            object_state = "blue"

    # if object_state is not red or blue but isFuture is True, then object_state is orange
    if object_state not in ["red", "blue"]:
        if isFuture is True:
            object_state = "orange"

    # if object_state is red but isUnderMonitoringDays is True and isUnderMonitoringHours is False, then object_state is orange
    if object_state == "red":
        if isUnderMonitoringDays is False or isUnderMonitoringHours is False:
            object_state = "orange"

    # define anomaly_reason
    if object_state == "green":
        status_message.append(isUnderDelayMessage)
        status_message.append(isUnderLatencyMessage)
        anomaly_reason.append("none")

        # if in a logical group, add the logical group message
        if isUnderLogicalGroup is True:
            status_message.append(LogicalGroupMsg)

    else:
        if isOutlier == 1:
            outlier_reasons = record.get("isOutlierReason", [])
            if isinstance(outlier_reasons, list):
                # Join the list elements into a single string
                outlier_reasons_str = " | ".join(outlier_reasons)
                status_message.append(outlier_reasons_str)
            else:
                # If it's not a list, append it directly
                status_message.append(outlier_reasons)
            anomaly_reason.append("ml_outliers_detection")

        if isAnomaly == 1:
            status_message.append(
                "anomalies detected in the data sampling and format recognition, review the data sampling screen to investigate. This alert means that trackMe detected an issue in the format of the events compared to the format that was previously identified for this source"
            )
            anomaly_reason.append("data_sampling_anomaly")

        if isFuture is True:
            status_message.append(isFutureMsg)
            anomaly_reason.append("future_over_tolerance")

        # monitoring days and hours, add the message first then the anomaly reason
        if isUnderMonitoringDays is False or isUnderMonitoringHours is False:
            status_message.append(isUnderMonitoringMsg)

        if isUnderMonitoringDays is False:
            anomaly_reason.append("out_of_monitoring_days")

        if isUnderMonitoringHours is False:
            anomaly_reason.append("out_of_monitoring_hours")

        if isUnderDcountHost is True:
            status_message.append(isUnderDcountHostMsg)
            anomaly_reason.append("min_hosts_dcount")

        if isUnderLatencyAlert is True:
            status_message.append(isUnderLatencyMessage)
            anomaly_reason.append("lag_threshold_breached")

        if isUnderDelayAlert is True:
            status_message.append(isUnderDelayMessage)
            anomaly_reason.append("delay_threshold_breached")

        # logical group
        if isUnderLogicalGroup is True:
            status_message.append(LogicalGroupMsg)
            anomaly_reason.append("in_logical_group")

    # form status_message_json
    status_message_json["status_message"] = status_message
    status_message_json["anomaly_reason"] = anomaly_reason

    # return
    return object_state, status_message, status_message_json, anomaly_reason


def set_dhm_status(
    splunkd_uri,
    session_key,
    tenant_id,
    record,
    isOutlier,
    isFuture,
    isFutureMsg,
    isUnderMonitoringDays,
    isUnderMonitoringHours,
    isUnderMonitoringMsg,
    object_logical_group_dict,
    isUnderLatencyAlert,
    isUnderLatencyMessage,
    isUnderDelayAlert,
    isUnderDelayMessage,
    default_splk_dhm_alerting_policy,
    data_lag_alert_kpis,
):
    """
    Create a function called set_dhm_status:
    - arguments: record, isOutlier, isFuture, isFutureMsg, isUnderMonitoringDays, isUnderMonitoringHours, object_logical_group_dict, isUnderLatencyAlert, isUnderLatencyMessage, isUnderDelayAlert, isUnderDelayMessage, default_splk_dhm_alerting_policy, data_lag_alert_kpis
    - returns:
        object_state (string): blue, orange, green, red
        anomaly_reason (list): list of short code reasons why the object is in anomaly
        status_message (list): list of long description reasons why the object is in anomaly
    - behaviour:
        object_state:
            green if:
                isOutlier is 1
                isFuture is False
                isUnderMonitoringDays is True
                isUnderMonitoringHours is True
                if isUnderLogicalGroup is True, then LogicalGroupStateInAlert must be False
                isUnderLatencyAlert is False
                isUnderDelayAlert is False
            blue if:
                Any of the condition above is not met, but isUnderLogicalGroup is True and LogicalGroupStateInAlert is True
            orange if:
                All green conditions are met except for isFuture which would be True
            red if:
                Any of the green conditions are not met, and blue conditions and orange conditions are not met
        anomaly_reason:
            if object_state is green, anomnaly_reason is None
            Otherwise, anomaly_reason is a list containing the reasons why the object is in anomaly
    """

    # init status_message and anomaly_reason
    status_message = []
    anomaly_reason = []

    # init status_message_json
    status_message_json = {}

    # define object_state
    if (
        (isOutlier == 0 or isOutlier == 2)
        and isUnderLatencyAlert is False
        and isUnderDelayAlert is False
    ):
        object_state = "green"
    else:
        object_state = "red"

    #
    # Logical group management
    #

    (
        isUnderLogicalGroup,
        LogicalGroupStateInAlert,
        LogicalGroupMsg,
    ) = get_and_manage_logical_group_status(
        splunkd_uri,
        session_key,
        tenant_id,
        record.get("object"),
        object_state,
        record.get("object_group_key"),
        object_logical_group_dict,
    )

    # log debug
    logging.debug(
        f'function get_and_manage_logical_group_status: object="{record.get("object")}", object_state="{object_state}", object_group_key="{record.get("object_group_key")}", isUnderLogicalGroup="{isUnderLogicalGroup}", LogicalGroupStateInAlert="{LogicalGroupStateInAlert}", LogicalGroupMsg="{LogicalGroupMsg}"'
    )

    splk_dhm_alerting_policy = record.get("splk_dhm_alerting_policy", "global_policy")
    if not len(splk_dhm_alerting_policy) > 0:
        splk_dhm_alerting_policy = "global_policy"
    splk_dhm_st_summary = record.get("splk_dhm_st_summary")

    # get the entity global max delay allowed
    global_max_delay_allowed = int(
        round(float(record.get("data_max_delay_allowed", 0)), 0)
    )

    # get the entity global max lag allowed
    global_max_lag_allowed = int(round(float(record.get("data_max_lag_allowed", 0)), 0))

    # get the entity global last delay seen (data_last_lag_seen)
    global_last_event_lag = int(round(float(record.get("data_last_lag_seen", 0)), 0))

    # get the entity global last lag seen (data_last_ingestion_lag_seen)
    global_last_ingest_lag = int(
        round(float(record.get("data_last_ingestion_lag_seen", 0)), 0)
    )

    # Convert splk_dhm_st_summary to a list if it is a string
    if isinstance(splk_dhm_st_summary, str):
        splk_dhm_st_summary = [splk_dhm_st_summary]

    # counters
    count_red = 0
    count_green = 0
    sourcetypes_red_list = []

    # retrieve host_idx_blocklists, host_st_blocklists
    host_idx_blocklists = record.get("host_idx_blocklists", [])
    host_st_blocklists = record.get("host_st_blocklists", [])

    # if string, then turn into list from comma separated string
    if isinstance(host_idx_blocklists, str):
        host_idx_blocklists = host_idx_blocklists.split(",")
    if isinstance(host_st_blocklists, str):
        host_st_blocklists = host_st_blocklists.split(",")

    # splk_dhm_st_summary can actually be a list
    if isinstance(splk_dhm_st_summary, list):

        for item_str in splk_dhm_st_summary:
            dict_str = "{" + item_str + "}"
            dict_loaded = False
            dict_loading_error = []

            # try both options
            try:
                new_dict = ast.literal_eval(dict_str)
                dict_loaded = True
            except Exception as e:
                dict_loaded = False
                dict_loading_error.append(str(e))

            if not dict_loaded:
                try:
                    new_dict = json.loads(item_str)
                    dict_loaded = True
                except Exception as e:
                    dict_loaded = False
                    dict_loading_error.append(str(e))

            if dict_loaded:

                # handle blocklists
                new_dict = {
                    key: val
                    for key, val in new_dict.items()
                    if val["idx"] not in host_idx_blocklists
                    and val["st"] not in host_st_blocklists
                }

                # Iterate through the inner dictionaries
                for inner_dict in new_dict.values():

                    if inner_dict.get("state") == "red":
                        count_red += 1

                        max_lag_allowed = float(inner_dict.get("max_lag_allowed"))
                        max_delay_allowed = float(inner_dict.get("max_delay_allowed"))
                        last_ingest_lag = float(inner_dict.get("last_ingest_lag"))
                        last_event_lag = float(inner_dict.get("last_event_lag"))

                        if (
                            last_ingest_lag > max_lag_allowed
                            and "lag_threshold_breached" not in anomaly_reason
                        ):
                            anomaly_reason.append("lag_threshold_breached")
                            sourcetypes_red_list.append(
                                f'(idx: {inner_dict.get("idx")}, st: {inner_dict.get("st")}, anomaly_reason: lag_threshold_breached)'
                            )
                        if (
                            last_event_lag > max_delay_allowed
                            and "delay_threshold_breached" not in anomaly_reason
                        ):
                            anomaly_reason.append("delay_threshold_breached")
                            sourcetypes_red_list.append(
                                f'(idx: {inner_dict.get("idx")}, st: {inner_dict.get("st")}, anomaly_reason: delay_threshold_breached)'
                            )

                    elif inner_dict.get("state") == "green":
                        count_green += 1

            else:
                logging.error(
                    f"Error in processing item_str: {item_str}. Error: {dict_loading_error}"
                )

        logging.debug(
            f'object="{record.get("object")}", count_red={count_red}, count_green={count_green}'
        )

        # turn sourcetypes_red_list into a pipe separated string
        sourcetypes_red_list = "|".join(sourcetypes_red_list)

        # Decision making based on the counts of red and green states
        if splk_dhm_alerting_policy == "global_policy":
            if default_splk_dhm_alerting_policy == "track_per_host":
                # Use object_state as it is
                pass
            elif default_splk_dhm_alerting_policy == "track_per_sourcetype":
                if count_red > 0:
                    object_state = "red"
                    status_message.append(
                        f"One or more sourcetypes are in alert for this entity, and policy is set to track_per_sourcetype, sourcetypes in alert: {sourcetypes_red_list}"
                    )
                else:
                    # Use object_state as it is
                    pass
        elif splk_dhm_alerting_policy == "track_per_host":
            # Use object_state as it is
            pass
        elif splk_dhm_alerting_policy == "track_per_sourcetype":
            if count_red > 0:
                object_state = "red"
                status_message.append(
                    f"One or more sourcetypes are in alert for this entity, and policy is set to track_per_sourcetype, sourcetypes in alert: {sourcetypes_red_list}"
                )
            else:
                # Use object_state as it is
                pass
        else:
            # Use object_state as it is
            pass

    # if all sourcetypes are in alert, object_state is red or orange depending on the global max delay entity values
    if (
        count_green == 0
        and (global_last_event_lag >= global_max_delay_allowed)
        and (global_last_ingest_lag >= global_max_lag_allowed)
    ):
        object_state = "red"
        status_message.append(
            f"all sourcetypes are in alert for this entity, global entity max delay allowed is breached (max_delay_allowed: {global_max_delay_allowed} seconds, duration: {convert_seconds_to_duration(global_max_delay_allowed)}, last_event_lag: {global_last_event_lag} seconds, duration: {convert_seconds_to_duration(global_last_event_lag)}), global entity max lag allowed is breached (max_delay_allowed: {global_max_lag_allowed} seconds, duration: {convert_seconds_to_duration(global_max_lag_allowed)}), last_event_lag: {global_last_ingest_lag} seconds, duration: {convert_seconds_to_duration(global_last_ingest_lag)})"
        )

    elif (
        count_green == 0
        and (global_last_event_lag < global_max_delay_allowed)
        and (global_last_ingest_lag >= global_max_lag_allowed)
    ):
        object_state = "red"
        status_message.append(
            f"all sourcetypes are in alert for this entity, global entity max delay allowed is not breached but max lag allowed is breached (max_delay_allowed: {global_max_delay_allowed} seconds, duration: {convert_seconds_to_duration(global_max_delay_allowed)}, last_event_lag: {global_last_event_lag} seconds, duration: {convert_seconds_to_duration(global_last_event_lag)})"
        )

    elif (
        count_green == 0
        and (global_last_event_lag >= global_max_delay_allowed)
        and (global_last_ingest_lag < global_max_lag_allowed)
    ):
        object_state = "red"
        status_message.append(
            f"all sourcetypes are in alert for this entity, global entity max delay allowed is breached but max lag allowed is not breached (max_delay_allowed: {global_max_delay_allowed} seconds, duration: {convert_seconds_to_duration(global_max_delay_allowed)}, last_event_lag: {global_last_event_lag} seconds, duration: {convert_seconds_to_duration(global_last_event_lag)})"
        )

    elif (
        count_green == 0
        and (global_last_event_lag < global_max_delay_allowed)
        and (global_last_ingest_lag < global_max_lag_allowed)
    ):
        object_state = "green"
        status_message.append(
            f"all sourcetypes are in alert for this entity, however global entity max delay allowed and max lag allowed are not breached (max_delay_allowed: {global_max_delay_allowed} seconds, duration: {convert_seconds_to_duration(global_max_delay_allowed)}, last_event_lag: {global_last_event_lag} seconds, duration: {convert_seconds_to_duration(global_last_event_lag)})"
        )

    elif count_green == 0:
        object_state = "red"
        status_message.append(
            f"all sourcetypes are in alert for this entity, however global entity level max delay allowed and max lag allowed could not be determined, verify TrackMe logs for more information (max_delay_allowed: {global_max_delay_allowed} seconds, duration: {convert_seconds_to_duration(global_max_delay_allowed)}, last_event_lag: {global_last_event_lag} seconds, duration: {convert_seconds_to_duration(global_last_event_lag)})"
        )

    # data_lag_alert_kpis set to lag_event_kpi: this overrides previously set object_state eventually
    if (
        object_state == "red"
        and (global_max_delay_allowed > global_last_event_lag)
        and data_lag_alert_kpis == "lag_event_kpi"
    ):
        object_state = "green"
        status_message.append(
            f"entity Alert over KPIs is set to {data_lag_alert_kpis} and global delay is not breached (max_delay_allowed: {global_max_delay_allowed} seconds, duration: {convert_seconds_to_duration(global_max_delay_allowed)})"
        )

    # data_lag_alert_kpis set to lag_ingestion_kpi: this overrides previously set object_state eventually
    if (
        object_state == "red"
        and (global_max_lag_allowed > global_last_ingest_lag)
        and data_lag_alert_kpis == "lag_ingestion_kpi"
    ):
        object_state = "green"
        status_message.append(
            f"entity Alert over KPIs is set to {data_lag_alert_kpis} and global lag is not breached (max_lag_allowed: {global_max_lag_allowed} seconds, duration: {convert_seconds_to_duration(global_max_lag_allowed)})"
        )

    # if object_state is red but isUnderLogicalGroup is True and LogicalGroupStateInAlert is False, then object_state is blue
    if object_state == "red" and isUnderLogicalGroup is True:
        if LogicalGroupStateInAlert is False:
            object_state = "blue"

    # if object_state is not red or blue but isFuture is True, then object_state is orange
    if object_state not in ["red", "blue"]:
        if isFuture is True:
            object_state = "orange"

    # if object_state is red but isUnderMonitoringDays is True and isUnderMonitoringHours is False, then object_state is orange
    if object_state == "red":
        if isUnderMonitoringDays is False or isUnderMonitoringHours is False:
            object_state = "orange"

    # define anomaly_reason
    if object_state == "green":
        status_message.append(isUnderDelayMessage)
        status_message.append(isUnderLatencyMessage)
        anomaly_reason.append("none")

        # if in a logical group, add the logical group message
        if isUnderLogicalGroup is True:
            status_message.append(LogicalGroupMsg)

    else:
        if isOutlier == 1:
            outlier_reasons = record.get("isOutlierReason", [])
            if isinstance(outlier_reasons, list):
                # Join the list elements into a single string
                outlier_reasons_str = " | ".join(outlier_reasons)
                status_message.append(outlier_reasons_str)
            else:
                # If it's not a list, append it directly
                status_message.append(outlier_reasons)
            anomaly_reason.append("ml_outliers_detection")

        if isFuture is True:
            status_message.append(isFutureMsg)
            anomaly_reason.append("future_over_tolerance")

        # monitoring days and hours, add the message first then the anomaly reason
        if isUnderMonitoringDays is False or isUnderMonitoringHours is False:
            status_message.append(isUnderMonitoringMsg)

        if isUnderMonitoringDays is False:
            anomaly_reason.append("out_of_monitoring_days")

        if isUnderMonitoringHours is False:
            anomaly_reason.append("out_of_monitoring_hours")

        if isUnderLatencyAlert is True:
            status_message.append(isUnderLatencyMessage)
            anomaly_reason.append("lag_threshold_breached")

        if isUnderDelayAlert is True:
            status_message.append(isUnderDelayMessage)
            anomaly_reason.append("delay_threshold_breached")

        # logical group
        if isUnderLogicalGroup is True:
            status_message.append(LogicalGroupMsg)
            anomaly_reason.append("in_logical_group")

    # form status_message_json
    status_message_json["status_message"] = status_message
    # deduplicate anomaly_reason
    anomaly_reason = list(set(anomaly_reason))
    status_message_json["anomaly_reason"] = anomaly_reason

    # return
    logging.debug(
        f'set_dhm_status, object="{record.get("object")}", object_state="{object_state}", status_message="{status_message}", anomaly_reason="{anomaly_reason}"'
    )
    return (
        object_state,
        status_message,
        status_message_json,
        anomaly_reason,
        splk_dhm_alerting_policy,
    )


def set_mhm_status(
    splunkd_uri,
    session_key,
    tenant_id,
    record,
    metric_details,
    isFuture,
    isFutureMsg,
    object_logical_group_dict,
):
    """
    Create a function called set_mhm_status:
    - arguments: record, isFuture, isFutureMsg, isUnderLogicalGroup, LogicalGroupStateInAlert
    - returns:
        object_state (string): blue, orange, green, red
        anomaly_reason (list): list of short code reasons why the object is in anomaly
        status_message (list): list of long description reasons why the object is in anomaly
    - behaviour:
        object_state:
            green if:
                all metric caterogies are green
            blue if:
                Any of the condition above is not met, but isUnderLogicalGroup is True and LogicalGroupStateInAlert is True
            orange if:
                All green conditions are met except for isFuture which would be True
            red if:
                Any of the green conditions are not met, and blue conditions and orange conditions are not met
        anomaly_reason:
            if object_state is green, anomnaly_reason is None
            Otherwise, anomaly_reason is a list containing the reasons why the object is in anomaly
    """

    # init status_message and anomaly_reason
    status_message = []
    anomaly_reason = []

    # init status_message_json
    status_message_json = {}

    # define object_state
    object_state = "green"

    #
    # Logical group management
    #

    (
        isUnderLogicalGroup,
        LogicalGroupStateInAlert,
        LogicalGroupMsg,
    ) = get_and_manage_logical_group_status(
        splunkd_uri,
        session_key,
        tenant_id,
        record.get("object"),
        object_state,
        record.get("object_group_key"),
        object_logical_group_dict,
    )

    # log debug
    logging.debug(
        f'function get_and_manage_logical_group_status: object="{record.get("object")}", object_state="{object_state}", object_group_key="{record.get("object_group_key")}", isUnderLogicalGroup="{isUnderLogicalGroup}", LogicalGroupStateInAlert="{LogicalGroupStateInAlert}", LogicalGroupMsg="{LogicalGroupMsg}"'
    )

    # Convert metric_details to a list if it is a string
    if isinstance(metric_details, str):
        metric_details = [metric_details]

    # counters
    count_red = 0
    count_green = 0
    metrics_red_list = []

    # splk_dhm_st_summary can actually be a list
    if isinstance(metric_details, list):
        for item_str in metric_details:
            try:
                new_dict = ast.literal_eval(item_str)

                # Iterate through the inner dictionaries
                for inner_dict in new_dict.values():
                    if inner_dict.get("state") == "red":
                        count_red += 1
                        anomaly_reason.append("delay_threshold_breached")
                        metrics_red_list.append(
                            f'(idx: {inner_dict.get("idx")}, metrics: {inner_dict.get("metric_category")}, anomaly_reason: delay_threshold_breached)'
                        )

                    elif inner_dict.get("state") == "green":
                        count_green += 1

            except Exception as e:
                logging.error(
                    f"Error in processing item_str: {item_str}. Error: {str(e)}"
                )

        logging.debug(
            f'object="{record.get("object")}", count_red={count_red}, count_green={count_green}'
        )

        # turn metrics_red_list into a pipe separated string
        metrics_red_list = "|".join(metrics_red_list)

        # Decision making based on the counts of red and green states
        if count_red > 0:
            object_state = "red"
            status_message.append(
                f"One or more metric categories are in alert for this entity, metrics in alert: {metrics_red_list}"
            )
        else:
            # Use object_state as it is
            pass

    # if all metrics are in alert, then object_state is red
    if count_green == 0:
        object_state = "red"
        status_message.append("all metric categories are in alert for this entity")

    # if object_state is red but isUnderLogicalGroup is True and LogicalGroupStateInAlert is False, then object_state is blue
    if object_state == "red" and isUnderLogicalGroup is True:
        if LogicalGroupStateInAlert is False:
            object_state = "blue"

    # if object_state is not red or blue but isFuture is True, then object_state is orange
    if object_state not in ["red", "blue"]:
        if isFuture is True:
            object_state = "orange"

    # define anomaly_reason
    if object_state == "green":
        status_message.append(
            "All metric categories are in normal state for this entity"
        )
        anomaly_reason.append("none")

        # if in a logical group, add the logical group message
        if isUnderLogicalGroup is True:
            status_message.append(LogicalGroupMsg)

    else:
        if isFuture is True:
            status_message.append(isFutureMsg)
            anomaly_reason.append("future_over_tolerance")

        # logical group
        if isUnderLogicalGroup is True:
            status_message.append(LogicalGroupMsg)
            anomaly_reason.append("in_logical_group")

    # form status_message_json
    status_message_json["status_message"] = status_message
    status_message_json["anomaly_reason"] = anomaly_reason

    # return
    logging.debug(
        f'set_mhm_status, object="{record.get("object")}", object_state="{object_state}", status_message="{status_message}", anomaly_reason="{anomaly_reason}"'
    )
    return object_state, status_message, status_message_json, anomaly_reason


def set_flx_status(
    splunkd_uri,
    session_key,
    tenant_id,
    record,
    isOutlier,
    isUnderMonitoringDays,
    isUnderMonitoringHours,
    isUnderMonitoringMsg,
    object_logical_group_dict,
):
    """
    Create a function called set_flx_status:
    - arguments: record, isOutlier, isFuture, isUnderMonitoringDays, isUnderMonitoringHours, isUnderLogicalGroup, LogicalGroupStateInAlert, isUnderLatencyAlert, isUnderLatencyMessage, isUnderDelayAlert, isUnderDelayMessage
    - returns:
        object_state (string): blue, orange, green, red
        anomaly_reason (list): list of short code reasons why the object is in anomaly
        status_message (list): list of long description reasons why the object is in anomaly
    - behaviour:
        object_state:
            green if:
                isOutlier is 1
                isFuture is False
                isUnderMonitoringDays is True
                isUnderMonitoringHours is True
                if isUnderLogicalGroup is True, then LogicalGroupStateInAlert must be False
                isUnderLatencyAlert is False
                isUnderDelayAlert is False
            blue if:
                Any of the condition above is not met, but isUnderLogicalGroup is True and LogicalGroupStateInAlert is True
            orange if:
                All green conditions are met except for isFuture which would be True
            red if:
                Any of the green conditions are not met, and blue conditions and orange conditions are not met
        anomaly_reason:
            if object_state is green, anomnaly_reason is None
            Otherwise, anomaly_reason is a list containing the reasons why the object is in anomaly
    """

    # init status_message and anomaly_reason
    status_message = []
    anomaly_reason = []

    # init status_message_json
    status_message_json = {}

    # for flx, attempt to retrieve extra_attributes, if present attempt to load as an object
    extra_attributes = record.get("extra_attributes", {})
    if isinstance(extra_attributes, str):
        if len(extra_attributes) > 0:
            try:
                extra_attributes = json.loads(extra_attributes)
            except Exception as e:
                logging.error(
                    f"Error in processing extra_attributes: {extra_attributes}. Error: {str(e)}"
                )
        else:
            extra_attributes = {}

    # for flx, first retrieve object_state which is defined upstream
    object_state = record.get("object_state", "green")

    #
    # Logical group management
    #

    (
        isUnderLogicalGroup,
        LogicalGroupStateInAlert,
        LogicalGroupMsg,
    ) = get_and_manage_logical_group_status(
        splunkd_uri,
        session_key,
        tenant_id,
        record.get("object"),
        object_state,
        record.get("object_group_key"),
        object_logical_group_dict,
    )

    # log debug
    logging.debug(
        f'function get_and_manage_logical_group_status: object="{record.get("object")}", object_state="{object_state}", object_group_key="{record.get("object_group_key")}", isUnderLogicalGroup="{isUnderLogicalGroup}", LogicalGroupStateInAlert="{LogicalGroupStateInAlert}", LogicalGroupMsg="{LogicalGroupMsg}"'
    )

    # status and status_description are used to compose the anomaly_reason
    status = record.get("status", "unknown")
    try:
        status = int(status)
    except Exception as e:
        pass
    status_description = record.get("status_description", "unknown")

    # Verify isOutlier
    if isOutlier == 1:
        object_state = "red"
    else:
        pass

    # If an Outlier was previously detected, but then fixed, the object_state will have switched to red
    # However, if the entity is not actively managed by a tracker, it wouldnt have been updated to green, we need to manage it here
    if object_state == "red" and status == 1 and isOutlier == 0:
        object_state = "green"

    # if object_state is red but isUnderMonitoringDays is True and isUnderMonitoringHours is False, then object_state is orange
    if object_state == "red":
        if isUnderMonitoringDays is False or isUnderMonitoringHours is False:
            object_state = "orange"

    # define anomaly_reason
    if object_state == "green":
        status_message_str = f"The entity status is complying with monitoring rules (status: {status}, status_description: {status_description})"
        status_message.append(status_message_str)
        anomaly_reason.append("none")

        # if in a logical group, add the logical group message
        if isUnderLogicalGroup is True:
            status_message.append(LogicalGroupMsg)

    else:
        # in flx, the status is managed upstream via the flag field status, unless red due to inactivity
        if status != 1:

            # get max_sec_inactive
            max_sec_inactive = record.get("max_sec_inactive", 0)
            try:
                max_sec_inactive = int(max_sec_inactive)
            except Exception as e:
                max_sec_inactive = 0

            # get the age in seconds since the latest execution
            sec_since_last_execution = round(
                time.time() - float(record.get("tracker_runtime")), 0
            )
            duration_since_last_execution = convert_seconds_to_duration(
                sec_since_last_execution
            )

            # Check and act
            if float(sec_since_last_execution) > max_sec_inactive:
                status_message_str = f"This entity has been inactive for more than {duration_since_last_execution} (D+HH:MM:SS) and was not actively managed by any tracker, its status was updated automatically by the inactive entities tracker"
                status_message = [status_message_str]
                status_description_short = "entity is red due to inactivity"
                status_description = f"The entity status is red due to inactivity, it was not actively managed by any tracker for more than {duration_since_last_execution} (D+HH:MM:SS)"
                anomaly_reason = ["inactive"]
                # in this case, we need to update the status_description and status_description_short
                record["status_description"] = status_description
                record["status_description_short"] = status_description_short

            else:
                status_message_str = f"The entity status is not complying with monitoring rules (status: {status}, status_description: {status_description})"
                status_message.append(status_message_str)
                anomaly_reason.append("status_not_met")

        # Other statements
        if isOutlier == 1:
            outlier_reasons = record.get("isOutlierReason", [])
            if isinstance(outlier_reasons, list):
                # Join the list elements into a single string
                outlier_reasons_str = " | ".join(outlier_reasons)
                status_message.append(outlier_reasons_str)
            else:
                # If it's not a list, append it directly
                status_message.append(outlier_reasons)
            anomaly_reason.append("ml_outliers_detection")

        # monitoring days and hours, add the message first then the anomaly reason
        if isUnderMonitoringDays is False or isUnderMonitoringHours is False:
            status_message.append(isUnderMonitoringMsg)

        if isUnderMonitoringDays is False:
            anomaly_reason.append("out_of_monitoring_days")

        if isUnderMonitoringHours is False:
            anomaly_reason.append("out_of_monitoring_hours")

        # logical group
        if isUnderLogicalGroup is True:
            status_message.append(LogicalGroupMsg)
            anomaly_reason.append("in_logical_group")

    # form status_message_json
    status_message_json["status_message"] = status_message
    status_message_json["anomaly_reason"] = anomaly_reason
    if extra_attributes:
        status_message_json["extra_attributes"] = extra_attributes

    # return
    logging.debug(
        f'set_dhm_status, object="{record.get("object")}", object_state="{object_state}", status_message="{status_message}", status_message_json="{status_message_json}", anomaly_reason="{anomaly_reason}"'
    )
    return object_state, status_message, status_message_json, anomaly_reason


def set_wlk_status(
    splunkd_uri,
    session_key,
    tenant_id,
    record,
    isOutlier,
    isUnderMonitoringDays,
    isUnderMonitoringHours,
    isUnderMonitoringMsg,
):
    """
    Create a function called set_wlk_status:
    - arguments: record, isOutlier, isFuture, isUnderMonitoringDays, isUnderMonitoringHours, isUnderLogicalGroup, LogicalGroupStateInAlert, isUnderLatencyAlert, isUnderLatencyMessage, isUnderDelayAlert, isUnderDelayMessage
    - returns:
        object_state (string): blue, orange, green, red
        anomaly_reason (list): list of short code reasons why the object is in anomaly
        status_message (list): list of long description reasons why the object is in anomaly
    - behaviour:
        object_state:
            green if:
                isOutlier is 1
                isFuture is False
                isUnderMonitoringDays is True
                isUnderMonitoringHours is True
                if isUnderLogicalGroup is True, then LogicalGroupStateInAlert must be False
                isUnderLatencyAlert is False
                isUnderDelayAlert is False
            blue if:
                Any of the condition above is not met, but isUnderLogicalGroup is True and LogicalGroupStateInAlert is True
            orange if:
                All green conditions are met except for isFuture which would be True
            red if:
                Any of the green conditions are not met, and blue conditions and orange conditions are not met
        anomaly_reason:
            if object_state is green, anomnaly_reason is None
            Otherwise, anomaly_reason is a list containing the reasons why the object is in anomaly
    """

    # init status_message and anomaly_reason
    status_message = []
    anomaly_reason = []

    # init status_message_json
    status_message_json = {}

    # for wlk, first retrieve object_state which is defined upstream
    object_state = record.get("object_state", "green")

    # status and status_description are used to compose the anomaly_reason
    status = record.get("status", "unknown")
    status_description = record.get("status_description", "unknown")

    # Verify isOutlier
    if isOutlier == 1:
        object_state = "red"
    else:
        pass

    # for wlk, get various functional fields used for the anomaly_reason and status_message definition
    # skipping KPis: skipped_pct, skipped_pct_last_60m, skipped_pct_last_4h, skipped_pct_last_24h
    try:
        skipped_pct = float(record.get("skipped_pct", 0))
    except Exception as e:
        skipped_pct = 0

    try:
        skipped_pct_last_60m = float(record.get("skipped_pct_last_60m", 0))
    except Exception as e:
        skipped_pct_last_60m = 0

    try:
        skipped_pct_last_4h = float(record.get("skipped_pct_last_4h", 0))
    except Exception as e:
        skipped_pct_last_4h = 0

    try:
        skipped_pct_last_24h = float(record.get("skipped_pct_last_24h", 0))
    except Exception as e:
        skipped_pct_last_24h = 0

    # similarly, load:
    # count_errors, count_errors_last_60m, count_errors_last_4h, count_errors_last_24h
    try:
        count_errors = int(record.get("count_errors", 0))
    except Exception as e:
        count_errors = 0

    try:
        count_errors_last_60m = int(record.get("count_errors_last_60m", 0))
    except Exception as e:
        count_errors_last_60m = 0

    try:
        count_errors_last_4h = int(record.get("count_errors_last_4h", 0))
    except Exception as e:
        count_errors_last_4h = 0

    try:
        count_errors_last_24h = int(record.get("count_errors_last_24h", 0))
    except Exception as e:
        count_errors_last_24h = 0

    # retrieve last_seen (epochtime) and cron_exec_sequence_sec (value in seconds)
    try:
        last_seen = int(record.get("last_seen", 0))
    except Exception as e:
        last_seen = 0

    # get last_seen_datetime
    if last_seen > 0:
        last_seen_datetime = convert_epoch_to_datetime(last_seen)
    else:
        last_seen_datetime = "unknown"

    try:
        cron_exec_sequence_sec = int(record.get("cron_exec_sequence_sec", 0))
    except Exception as e:
        cron_exec_sequence_sec = 0

    # calculate isDelayed (0 or 1)
    # if now()-last_seen)>(cron_exec_sequence_sec+3600, isDelayed is 1
    now = time.time()
    if (now - last_seen) > (cron_exec_sequence_sec + 3600):
        isDelayed = 1
    else:
        isDelayed = 0

    # calculate the current delay in seconds
    current_delay = now - last_seen

    # get the current delay durection
    current_delay_duration = convert_seconds_to_duration(current_delay)

    # retrieve orphan boolean (0 or 1) and load as an integer, as well as orphan_last_check (human readable date)
    try:
        orphan = int(record.get("orphan", 0))
    except Exception as e:
        orphan = 0

    orphan_last_check = record.get("orphan_last_check", "unknown")

    # if object_state is red but isUnderMonitoringDays is True and isUnderMonitoringHours is False, then object_state is orange
    if object_state == "red":
        if isUnderMonitoringDays is False or isUnderMonitoringHours is False:
            object_state = "orange"

    # define anomaly_reason
    if object_state == "green":
        status_message_str = f"The entity status is complying with monitoring rules (status: {status}, status_description: {status_description})"
        status_message.append(status_message_str)
        anomaly_reason.append("none")

    else:
        # Other statements

        # ML Outlier
        if isOutlier == 1:
            outlier_reasons = record.get("isOutlierReason", [])
            if isinstance(outlier_reasons, list):
                # Join the list elements into a single string
                outlier_reasons_str = " | ".join(outlier_reasons)
                status_message.append(outlier_reasons_str)
            else:
                # If it's not a list, append it directly
                status_message.append(outlier_reasons)
            anomaly_reason.append("ml_outliers_detection")

        # Skipping
        if (
            skipped_pct > 0
            or skipped_pct_last_60m > 0
            or skipped_pct_last_4h > 0
            or skipped_pct_last_24h > 0
        ):
            status_message.append(
                f"skipping searches were detected, review and address performance issues for this search or finetune its scheduling plan to clear this alert. (skipped_pct_last_60m: {skipped_pct_last_60m}, skipped_pct_last_4h: {skipped_pct_last_4h}, skipped_pct_last_24h: {skipped_pct_last_24h})"
            )
            anomaly_reason.append("skipping_searches_detected")

        # Errors count
        if (
            count_errors > 0
            or count_errors_last_60m > 0
            or count_errors_last_4h > 0
            or count_errors_last_24h > 0
        ):
            status_message.append(
                f"execution errors were detected, review and address these errors to clear this alert. (count_errors_last_60m: {count_errors_last_60m}, count_errors_last_4h: {count_errors_last_4h}, count_errors_last_24h: {count_errors_last_24h})"
            )
            anomaly_reason.append("execution_errors_detected")

        # orphan
        if orphan == 1:
            status_message.append(
                f"the search was detected as an orphan search which means the user owning the search is not currently a valid user (orphan: {orphan}, time check: {orphan_last_check}"
            )
            anomaly_reason.append("orphan_search_detected")

        # delayed
        if isDelayed == 1 and cron_exec_sequence_sec > 0:
            status_message.append(
                f"the search was detected as delayed, this means the search is not running as expected (isDelayed: {isDelayed}, last_seen: {last_seen_datetime}, cron_exec_sequence_sec: {cron_exec_sequence_sec}, current delay: {current_delay_duration} duration)"
            )
            anomaly_reason.append("execution_delayed")

        # monitoring days and hours, add the message first then the anomaly reason
        if isUnderMonitoringDays is False or isUnderMonitoringHours is False:
            status_message.append(isUnderMonitoringMsg)

        if isUnderMonitoringDays is False:
            anomaly_reason.append("out_of_monitoring_days")

        if isUnderMonitoringHours is False:
            anomaly_reason.append("out_of_monitoring_hours")

        # if we failed to identify the reason
        if len(status_message) == 0:
            status_message_str = f"The entity status is not complying with monitoring rules (status: {status}, status_description: {status_description})"
            status_message.append(status_message_str)
            anomaly_reason.append("status_not_met")

    # form status_message_json
    status_message_json["status_message"] = status_message
    status_message_json["anomaly_reason"] = anomaly_reason

    # return
    logging.debug(
        f'set_wlk_status, object="{record.get("object")}", object_state="{object_state}", status_message="{status_message}", anomaly_reason="{anomaly_reason}"'
    )
    return object_state, status_message, status_message_json, anomaly_reason


def set_cim_status(
    splunkd_uri,
    session_key,
    tenant_id,
    record,
    isOutlier,
    isUnderMonitoringDays,
    isUnderMonitoringHours,
    isUnderMonitoringMsg,
):
    """
    Create a function called set_cim_status:
    - arguments: record, isOutlier
    - returns:
        object_state (string): blue, orange, green, red
        anomaly_reason (list): list of short code reasons why the object is in anomaly
        status_message (list): list of long description reasons why the object is in anomaly
    - behaviour:
        object_state:
            green if:
                isOutlier is 1
                cim compliance is False
            blue if:
                Any of the condition above is not met, but isUnderLogicalGroup is True and LogicalGroupStateInAlert is True
            orange if:
                All green conditions are met except for isFuture which would be True
            red if:
                Any of the green conditions are not met, and blue conditions and orange conditions are not met
        anomaly_reason:
            if object_state is green, anomnaly_reason is None
            Otherwise, anomaly_reason is a list containing the reasons why the object is in anomaly
    """

    # init status_message and anomaly_reason
    status_message = []
    anomaly_reason = []

    # init status_message_json
    status_message_json = {}

    # for cim, attempt to retrieve cim_tracking_rules and cim_tracking_results, if present attempt to load as an object
    cim_tracking_rules = record.get("cim_tracking_rules", {})
    if isinstance(cim_tracking_rules, str):
        try:
            cim_tracking_rules = json.loads(cim_tracking_rules)
        except Exception as e:
            logging.error(
                f"Error in processing cim_tracking_rules: {cim_tracking_rules}. Error: {str(e)}"
            )

    cim_tracking_results = record.get("cim_tracking_results", {})
    if isinstance(cim_tracking_results, str):
        try:
            cim_tracking_results = json.loads(cim_tracking_results)
        except Exception as e:
            logging.error(
                f"Error in processing cim_tracking_results: {cim_tracking_results}. Error: {str(e)}"
            )

    # for cim, first retrieve object_state which is defined upstream
    object_state = record.get("object_state", "red")
    if object_state == "unknown":  # if returned unknown, then it's red
        object_state = "red"

    # Verify isOutlier
    if isOutlier == 1:
        object_state = "red"
    else:
        pass

    # if object_state is red but isUnderMonitoringDays is True and isUnderMonitoringHours is False, then object_state is orange
    if object_state == "red":
        if isUnderMonitoringDays is False or isUnderMonitoringHours is False:
            object_state = "orange"

    # define anomaly_reason
    if object_state == "green":
        status_message_str = (
            f"all compliancy checks on the CIM entity are currently successful."
        )
        status_message.append(status_message_str)
        anomaly_reason.append("none")

    else:
        # in cim, the anomaly_reason is part of the cim_tracking_results

        if len(cim_tracking_results) > 0:
            try:
                anomaly_reason_str = cim_tracking_results.get("anomaly_reason", None)

                if anomaly_reason_str:
                    """
                    if anomaly_reason_str contains:
                    - "compliance_failed:", anomaly_reason is "compliance_failed", status_message is:
                    "CIM compliancy checks have failed for one or more CIM fields: [ " . cim_fields_red . " ], review the compliancy check results and rules, idenfity and address the root causes to fix the entity state."
                    - "no_results:", anomaly_reason is "no_results", status_message is:
                    "Alert: CIM compliancy checks have failed since the main search did not return any results, review the entity definition and its schedule."
                    """

                    if anomaly_reason_str.startswith("compliance_failed:"):
                        anomaly_reason_str = "compliance_failed"
                        cim_fields_red = cim_tracking_results.get(
                            "cim_fields_red", None
                        )
                        if cim_fields_red:
                            status_message_str = f"CIM compliancy checks have failed for one or more CIM fields: {cim_fields_red}, review the compliancy check results and rules, idenfity and address the root causes to fix the entity state."
                        else:
                            status_message_str = f"CIM compliancy checks have failed for compliance issue reason but we failed to extract the cim_fields_red!"

                    elif anomaly_reason_str.startswith("no_results:"):
                        anomaly_reason_str = "no_results"
                        status_message_str = f"CIM compliancy checks have failed since the main search did not return any results, review the entity definition and its schedule."

                    anomaly_reason.append(anomaly_reason_str)
                    status_message.append(status_message_str)

                else:
                    anomaly_reason_str = "unknown"
                    status_message_str = (
                        "CIM compliancy checks have failed for an unknown reason!"
                    )
                    anomaly_reason.append(anomaly_reason_str)
                    status_message.append(status_message_str)

            except Exception as e:
                logging.error(
                    f'failed to process the record="{json.dumps(record, indent=2)}", exception="{str(e)}"'
                )
                anomaly_reason_str = "unknown"
                status_message_str = (
                    "CIM compliancy checks have failed for an unknown reason!"
                )
                anomaly_reason.append(anomaly_reason_str)
                status_message.append(status_message_str)

        else:
            anomaly_reason_str = "unknown"
            status_message_str = "CIM compliancy checks have not been performed yet, or no results were resulting since the CIM entity was created, run the tracker now and review the results again."
            anomaly_reason.append(anomaly_reason_str)
            status_message.append(status_message_str)

        # Other statements
        if isOutlier == 1:
            outlier_reasons = record.get("isOutlierReason", [])
            if isinstance(outlier_reasons, list):
                # Join the list elements into a single string
                outlier_reasons_str = " | ".join(outlier_reasons)
                status_message.append(outlier_reasons_str)
            else:
                # If it's not a list, append it directly
                status_message.append(outlier_reasons)
            anomaly_reason.append("ml_outliers_detection")

        # monitoring days and hours, add the message first then the anomaly reason
        if isUnderMonitoringDays is False or isUnderMonitoringHours is False:
            status_message.append(isUnderMonitoringMsg)

        if isUnderMonitoringDays is False:
            anomaly_reason.append("out_of_monitoring_days")

        if isUnderMonitoringHours is False:
            anomaly_reason.append("out_of_monitoring_hours")

    # form status_message_json
    status_message_json["status_message"] = status_message
    status_message_json["anomaly_reason"] = anomaly_reason

    # return
    logging.debug(
        f'set_cim_status, object="{record.get("object")}", object_state="{object_state}", status_message="{status_message}", status_message_json="{status_message_json}", anomaly_reason="{anomaly_reason}"'
    )
    return object_state, status_message, status_message_json, anomaly_reason


def ack_check(object_value, ack_collection_keys, ack_collection_dict, record):
    """
    Updates record with ack information if object_value exists in ack_collection_keys
    and the object categories match.
    """

    ack_defaults = {
        "ack_state": "inactive",
        "ack_type": "N/A",
        "ack_comment": "N/A",
        "ack_expiration": "N/A",
        "ack_mtime": "N/A",
    }

    if object_value in ack_collection_keys:
        ack_record = ack_collection_dict.get(object_value)
        # Check if ack_record exists and object_category matches
        if ack_record and ack_record.get("object_category") == record.get(
            "object_category"
        ):
            # Extract ack information from ack_record
            for key in ack_defaults.keys():
                record[key] = ack_record.get(key, ack_defaults[key])
        else:
            record.update(ack_defaults)
    else:
        record.update(ack_defaults)


def define_state_icon_code(record):
    """
    Determines the state_icon_code based on the object_state and ack_state
    contained within the record dictionary.

    Args:
    - record (dict): A dictionary containing 'object_state' and 'ack_state'.

    Returns:
    - str: The state_icon_code determined based on the provided conditions.
    """
    object_state = record.get("object_state")
    ack_state = record.get("ack_state")

    # Define a mapping based on the Splunk macro conditions
    state_icon_code_mapping = {
        ("green", "inactive"): "001",
        ("green", "active"): "002",
        ("green", None): "003",
        ("red", "inactive"): "004",
        ("red", "active"): "005",
        ("red", None): "006",
        ("orange", "inactive"): "007",
        ("orange", "active"): "008",
        ("orange", None): "009",
        ("blue", "inactive"): "010",
        ("blue", "active"): "011",
        ("blue", None): "012",
    }

    # Fallback code if none of the conditions match
    default_code = "999"

    # Determine state_icon_code based on object_state and ack_state
    state_icon_code = state_icon_code_mapping.get(
        (object_state, ack_state),
        state_icon_code_mapping.get((object_state, None), default_code),
    )

    return state_icon_code


def outliers_readiness(icons_state_mode, record):
    """
    Updates the record with outliers_readiness, OutliersIsOk, and OutliersStatus based on the record's values.
    Ensures robust handling of missing or non-integer isOutlier values.
    """
    # Set outliers_readiness based on its current value
    record["outliers_readiness"] = (
        "True" if record.get("outliers_readiness") == "True" else "False"
    )

    # Safely get and convert isOutlier to an integer
    try:
        is_outlier = int(record.get("isOutlier", 0))
    except ValueError:
        # Handle case where conversion fails
        is_outlier = 0

    # Determine OutliersIsOk based on isOutlier
    record["OutliersIsOk"] = 1 if is_outlier == 0 else 0

    # Set OutliersStatus based on OutliersIsOk
    if icons_state_mode == "emoji":
        record["OutliersStatus"] = "🟢" if record["OutliersIsOk"] == 1 else "🔴"
    else:
        record["OutliersStatus"] = "green" if record["OutliersIsOk"] == 1 else "red"


def sampling_anomaly_status(icons_state_mode, record):
    """
    Updates the record with SamplingIsOk and SamplingStatus based on the isAnomaly field.
    """

    # get isAnomaly
    try:
        isAnomaly = int(record.get("isAnomaly", 0))
    except Exception as e:
        isAnomaly = 0

    # define SamplingIsOk
    record["SamplingIsOk"] = 1 if isAnomaly == 0 else 1

    # define SamplingStatus
    if icons_state_mode == "emoji":
        record["SamplingStatus"] = "🟢" if record["SamplingIsOk"] == 1 else "🔴"
    else:
        record["SamplingStatus"] = "green" if record["SamplingIsOk"] == 1 else "red"


def logical_group_lookup(
    object_value,
    logicalgroup_members_collection_keys,
    logicalgroup_members_collection_dict,
    record,
):
    """
    Updates record with Logical Group information if object_value exists in lg_collection_keys.
    """

    logicalgroup_defaults = {
        "object_group_key": None,
        "object_group_name": None,
    }

    if object_value in logicalgroup_members_collection_keys:
        logicalgroup_record = logicalgroup_members_collection_dict.get(object_value)
        # Extract ack information from ack_record
        for key in logicalgroup_defaults.keys():
            record[key] = logicalgroup_record.get(key, logicalgroup_defaults[key])
    else:
        # for key in logicalgroup_defaults, remove the key from record if exists
        for key in logicalgroup_defaults.keys():
            if key in record:
                del record[key]


def set_feeds_lag_summary(record, component):
    """
    Generates a lag summary based on the data_last_lag_seen and data_last_ingestion_lag_seen fields
    """

    if component in ["dsm", "dhm"]:

        try:
            data_last_lag_seen = int(
                round(float(record.get("data_last_lag_seen", 0)), 0)
            )
        except Exception as e:
            data_last_lag_seen = 0
        try:
            data_last_ingestion_lag_seen = int(
                round(float(record.get("data_last_ingestion_lag_seen", 0)), 0)
            )
        except Exception as e:
            data_last_ingestion_lag_seen = 0

        if data_last_lag_seen > 60 or data_last_ingestion_lag_seen < -60:
            data_last_lag_seen_duration = (
                f"{convert_seconds_to_duration(data_last_lag_seen)}"
            )
        elif data_last_lag_seen == 0:
            data_last_lag_seen_duration = "0 sec"
        elif data_last_lag_seen < 60:
            data_last_lag_seen_duration = f"-{data_last_lag_seen} sec"
        else:
            data_last_lag_seen_duration = f"{data_last_lag_seen} sec"

        if data_last_ingestion_lag_seen > 60 or data_last_ingestion_lag_seen < -60:
            data_last_ingestion_lag_seen_duration = (
                f"{convert_seconds_to_duration(data_last_ingestion_lag_seen)}"
            )
        elif data_last_ingestion_lag_seen == 0:
            data_last_ingestion_lag_seen_duration = "0 sec"
        elif data_last_ingestion_lag_seen < 60:
            data_last_ingestion_lag_seen_duration = (
                f"-{data_last_ingestion_lag_seen} sec"
            )
        else:
            data_last_ingestion_lag_seen_duration = (
                f"{data_last_ingestion_lag_seen} sec"
            )

        # return
        lag_summary = (
            f"{data_last_lag_seen_duration} / {data_last_ingestion_lag_seen_duration}"
        )

    elif component in ["mhm"]:

        # original logic: lag_summary= if(last_lag_seen>60, tostring(last_lag_seen, "duration"), last_lag_seen . " sec")

        try:
            last_lag_seen = int(round(float(record.get("last_lag_seen", 0)), 0))
        except Exception as e:
            last_lag_seen = 0
        if last_lag_seen > 60:
            lag_summary = f"{convert_seconds_to_duration(last_lag_seen)}"
        else:
            lag_summary = f"{last_lag_seen} sec"

    return lag_summary


def set_feeds_thresholds_duration(record):

    try:
        data_max_delay_allowed = int(
            round(float(record.get("data_max_delay_allowed", 0)), 0)
        )
    except Exception as e:
        data_max_delay_allowed = 0
    data_max_delay_allowed_duration = convert_seconds_to_duration(
        data_max_delay_allowed
    )
    try:
        data_max_lag_allowed = int(
            round(float(record.get("data_max_lag_allowed", 0)), 0)
        )
    except Exception as e:
        data_max_lag_allowed = 0
    data_max_lag_allowed_duration = convert_seconds_to_duration(data_max_lag_allowed)

    return data_max_delay_allowed_duration, data_max_lag_allowed_duration


def set_cim_duration(record):

    try:
        tracker_last_duration = int(
            round(float(record.get("tracker_last_duration", 0)), 0)
        )
    except Exception as e:
        tracker_last_duration = 0
    tracker_last_duration = convert_seconds_to_duration(tracker_last_duration)

    return tracker_last_duration


def dsm_sampling_lookup(
    object_value, sampling_collection_keys, sampling_collection_dict, record
):
    """
    Updates record with ack information if object_value exists in sampling_collection_keys
    and the object categories match.
    """

    sampling_defaults = {
        "data_sample_feature": "enabled",
        "data_sample_status_message": {
            "state": "pending",
            "desc": "Data Sampling is pending and has not been performed yet for this entity",
        },
        "data_sample_status_colour": "N/A",
        "data_sample_anomaly_reason": "N/A",
    }

    if object_value in sampling_collection_keys:
        sampling_record = sampling_collection_dict.get(object_value)

        current_data_sample_feature = sampling_record.get(
            "data_sample_feature", "enabled"
        )

        if current_data_sample_feature == "enabled":
            for key in sampling_defaults.keys():
                record[key] = sampling_record.get(key, sampling_defaults[key])
        elif current_data_sample_feature == "disabled_auto":
            for key in sampling_defaults.keys():
                record[key] = sampling_record.get(key, sampling_defaults[key])
        else:
            sampling_fields = {
                "data_sample_feature": "disabled",
                "data_sample_status_message": {
                    "state": "disabled",
                    "desc": "Data sampling features are currently disabled for this entity.",
                },
                "data_sample_status_colour": "grey",
                "data_sample_anomaly_reason": "None",
            }
            for key in sampling_fields.keys():
                record[key] = sampling_fields[key]

    else:
        record.update(sampling_defaults)


def outliers_data_lookup(
    key_value,
    outliers_data_collection_keys,
    outliers_data_collection_dict,
    outliers_rules_collection_keys,
    outliers_rules_collection_dict,
    record,
):
    """
    Updates record with outliers information if object_value exists in outliers_data_collection_keys
    and the object categories match.
    """

    #
    # Handle data
    #

    outliers_data_defaults = {
        "isOutlier": 0,
        "isOutlierReason": "",
        "models_in_anomaly": "",
    }

    if key_value in outliers_data_collection_keys:
        outliers_data_record = outliers_data_collection_dict.get(key_value)
        for key in outliers_data_defaults.keys():
            record[key] = outliers_data_record.get(key, outliers_data_defaults[key])
    else:
        record.update(outliers_data_defaults)

    #
    # Handle rules
    #

    outliers_rules_defaults = {
        "is_disabled": 0,
    }

    if key_value in outliers_rules_collection_keys:
        outliers_rules_record = outliers_rules_collection_dict.get(key_value)

        for key in outliers_rules_defaults.keys():
            if key == "is_disabled":
                record["OutliersDisabled"] = outliers_rules_record.get(
                    key, outliers_rules_defaults[key]
                )
            else:
                record[key] = outliers_rules_record.get(
                    key, outliers_rules_defaults[key]
                )
    else:
        for key, value in outliers_rules_defaults.items():
            if key == "is_disabled":
                record["OutliersDisabled"] = value
            else:
                record[key] = value


def get_coll_docs_ref(collection, docs_collection_name):
    collection_records = []
    collection_records_dict = {}
    collection_members_list = []
    collection_members_dict = {}

    end = False
    skip_tracker = 0
    while not end:
        process_collection_records = collection.data.query(skip=skip_tracker)
        if process_collection_records:
            for item in process_collection_records:
                collection_records.append(item)
                collection_records_dict[item.get("_key")] = {
                    "doc_note": item.get("doc_note"),
                    "doc_link": item.get("doc_link"),
                    "object_members": item.get("object", []),
                }

                doc_members = item.get("object", [])
                # add members in collection_members_list, also create a dict per member
                for member in doc_members:
                    if member not in collection_members_list:
                        collection_members_list.append(member)
                        collection_members_dict[member] = {
                            "doc_key": item.get("_key"),
                            "doc_note": item.get("doc_note"),
                            "doc_link": item.get("doc_link"),
                            "object_members": item.get("object", []),
                        }
            skip_tracker += 5000
        else:
            end = True

    return (
        collection_records,
        collection_records_dict,
        collection_members_list,
        collection_members_dict,
    )


def docs_ref_lookup(
    docs_is_global,
    docs_note_default_global,
    docs_link_default_global,
    object_value,
    docs_members_collection_keys,
    docs_members_collection_dict,
    record,
):
    """
    Updates record with docs ref information if object_value exists in docs_members_collection_keys.
    """

    docs_defaults = {
        "doc_is_global": docs_is_global,
        "doc_note": docs_note_default_global,
        "doc_link": docs_link_default_global,
    }

    if object_value in docs_members_collection_keys:
        doc_record = docs_members_collection_dict.get(object_value)

        # override doc_is_global to False
        record["doc_is_global"] = False

        # Extract information from record
        for key in docs_defaults.keys():
            if key != "doc_is_global":
                record[key] = doc_record.get(key, docs_defaults[key])
    else:
        record.update(docs_defaults)


def wlk_disabled_apps_lookup(
    app_value,
    apps_enablement_collection_keys,
    apps_enablement_collection_dict,
    record,
):
    """
    Updates record with apps_disabled ref information if object_value exists in apps_enablement_collection_keys.
    """

    apps_enablement_defaults = {
        "app_is_enabled": "True",
    }

    if app_value in apps_enablement_collection_keys:
        lookup_record = apps_enablement_collection_dict.get(app_value)
        # Extract ack information from record
        record["app_is_enabled"] = lookup_record.get("enabled", "True")
    else:
        record.update(apps_enablement_defaults)


def wlk_versioning_lookup(
    key_value,
    versioning_collection_keys,
    versioning_collection_dict,
    record,
):
    """
    Updates record with apps_disabled ref information if object_value exists in cron_exec_sequence_sec_collection_keys.
    """

    versioning_defaults = {
        "cron_exec_sequence_sec": 0,
        "object_description": "No description",
        "versioning_available": "False",
    }

    # lookup and override if found, otherwise do nothing
    if key_value in versioning_collection_keys:
        record["versioning_available"] = "True"
        lookup_record = versioning_collection_dict.get(key_value)

        for key in versioning_defaults.keys():
            if key == "object_description":
                lookup_record_description = lookup_record.get("description", None)
                # if len of lookup_record_description is 0, then use the default
                if lookup_record_description:
                    if len(lookup_record_description) == 0:
                        record["object_description"] = versioning_defaults[
                            "object_description"
                        ]
                    else:
                        record["object_description"] = lookup_record_description
                else:
                    record["object_description"] = versioning_defaults[
                        "object_description"
                    ]

            elif key == "versioning_available":
                record["versioning_available"] = "True"
            else:
                record[key] = lookup_record.get(key, versioning_defaults[key])

        logging.debug(
            f'versioning found for object="{record.get("object")}", object_key="{record.get("keyid")}", using key_value="{key_value}"'
        )

    else:
        record.update(versioning_defaults)
        logging.debug(
            f'no versioning found for object="{record.get("object")}", object_key="{record.get("keyid")}", using key_value="{key_value}"'
        )


def wlk_orphan_lookup(
    key_value,
    orphan_collection_keys,
    orphan_collection_dict,
    record,
):
    """
    Updates record with orphan ref information if key_value exists in orphan_collection_keys.
    """

    orphan_defaults = {
        "orphan": 0,
    }

    if key_value in orphan_collection_keys:
        lookup_record = orphan_collection_dict.get(key_value)
        record["orphan"] = lookup_record.get("orphan", 0)
    else:
        record.update(orphan_defaults)


def apply_blocklist(record, blocklist_not_regex, blocklist_regex):
    """
    Determines whether a record should be appended based on blocklist rules.

    :param record: The record to check.
    :param blocklist_not_regex: Dict of blocklist rules without regex.
    :param blocklist_regex: Dict of blocklist rules with regex.
    :return: True if the record should be appended, False otherwise.
    """

    def match_not_regex(field_value, rule):
        """
        Check if a field value matches a non-regex blocklist rule.
        """
        if isinstance(field_value, list):
            return any(item == rule.get("object") for item in field_value)
        else:
            return field_value == rule.get("object")

    def match_regex(field_value, rule):
        """
        Check if a field value matches a regex blocklist rule.
        """
        if isinstance(field_value, list):
            return any(re.match(rule.get("object"), item) for item in field_value)
        else:
            return re.match(rule.get("object"), field_value)

    # define index and add to the record, using data_index if available and turn into a list from csv
    if "data_index" in record:
        record["index"] = record["data_index"].split(",")

    # same for data_sourcetype and sourcetype
    if "data_sourcetype" in record:
        record["sourcetype"] = record["data_sourcetype"].split(",")

    # metric_category is called equally in record and blocklist, but it can be a list too
    if "metric_category" in record:
        record["metric_category"] = record["metric_category"].split(",")

    # Check blocklist without regex
    for _, rule in blocklist_not_regex.items():
        object_category = rule.get("object_category")
        if object_category in record and match_not_regex(record[object_category], rule):
            return False  # Match found in blocklist not regex, do not append

    # Check blocklist with regex
    for _, rule in blocklist_regex.items():
        object_category = rule.get("object_category")
        if object_category in record and match_regex(record[object_category], rule):
            return False  # Regex match found in blocklist, do not append

    # before returning, remove index and sourcetype from the record
    if "index" in record:
        del record["index"]
    if "sourcetype" in record:
        del record["sourcetype"]
    # turn metric_category is existing back to a comma separated string
    if "metric_category" in record:
        record["metric_category"] = ",".join(record["metric_category"])

    return True  # If no blocklist rules matched, append the record


def dsm_check_default_thresholds(record, trackme_conf):
    """
    Verify that the record contains expected fields, if not or if they are None, set them to default values.
    """

    # Define a dictionary for the DSM fields and their respective default values
    fields_defaults = {
        "data_max_delay_allowed": trackme_conf["splk_general"][
            "splk_general_dsm_delay_default"
        ],
        "data_max_lag_allowed": trackme_conf["splk_general"][
            "splk_general_dsm_threshold_default"
        ],
        "data_lag_alert_kpis": "all_kpis",
        "data_override_lagging_class": "false",
        "allow_adaptive_delay": "true",
    }

    # Iterate through the fields to check their presence and value
    for field, default_value in fields_defaults.items():
        # Check if field is missing or explicitly set to None
        if field not in record or record[field] is None:
            record[field] = default_value
        else:
            # Additional checks for numeric fields to ensure they can be converted to float
            if field in ["data_max_delay_allowed", "data_max_lag_allowed"]:
                try:
                    # Attempt to convert to float to validate
                    record[field] = float(record[field])
                except ValueError:
                    # Set to default if conversion fails
                    record[field] = default_value


def dhm_check_default_thresholds(record, trackme_conf):
    """
    Verify that the record contains expected fields, if not or if they are None, set them to default values.
    """

    # Define a dictionary for the DHM fields and their respective default values
    fields_defaults = {
        "data_max_delay_allowed": trackme_conf["splk_general"][
            "splk_general_dhm_delay_default"
        ],
        "data_max_lag_allowed": trackme_conf["splk_general"][
            "splk_general_dhm_threshold_default"
        ],
        "data_lag_alert_kpis": "all_kpis",
        "data_override_lagging_class": "false",
        "allow_adaptive_delay": "true",
        "splk_dhm_alerting_policy": "global_policy",
    }

    # Iterate through the fields to check their presence and value
    for field, default_value in fields_defaults.items():
        # Check if field is missing or explicitly set to None
        if field not in record or record[field] is None:
            record[field] = default_value
        else:
            # Additional checks for numeric fields to ensure they can be converted to float
            if field in ["data_max_delay_allowed", "data_max_lag_allowed"]:
                try:
                    # Attempt to convert to float to validate
                    record[field] = float(record[field])
                except ValueError:
                    # Set to default if conversion fails
                    record[field] = default_value


def dynamic_priority_lookup(
    key_value, priority_collection_keys, priority_collection_dict, record
):
    """
    Updates record with dynamic priority information if key_value exists in priority_collection_keys.
    """

    if key_value in priority_collection_keys:
        dynamic_priority_record = priority_collection_dict.get(key_value)
        dynamic_priority = dynamic_priority_record.get("priority", None)

        # if we have a match, and a priority, then update the record, otherwise do nothing
        if dynamic_priority:
            record["priority"] = dynamic_priority
            logging.debug(
                f'match found applying dynamic priority="{dynamic_priority}" for object="{record.get("object")}", key="{key_value}"'
            )


def dynamic_tags_lookup(key_value, tags_collection_keys, tags_collection_dict, record):
    """
    Updates record with dynamic tags information if key_value exists in tags_collection_keys.
    """

    if key_value in tags_collection_keys:
        dynamic_tags_record = tags_collection_dict.get(key_value)
        dynamic_tags = dynamic_tags_record.get("tags_auto", None)

        # if we have a match, then update the record, otherwise do nothing
        if dynamic_tags:
            record["tags_auto"] = dynamic_tags
            logging.debug(
                f'match found applying dynamic tags="{dynamic_tags}" for object="{record.get("object")}", key="{key_value}"'
            )


def dynamic_sla_class_lookup(
    key_value, sla_collection_keys, sla_collection_dict, record
):
    """
    Updates record with dynamic priority information if key_value exists in priority_collection_keys.
    """

    if key_value in sla_collection_keys:
        dynamic_sla_record = sla_collection_dict.get(key_value)
        dynamic_sla = dynamic_sla_record.get("sla_class", None)

        # if we have a match, and a priority, then update the record, otherwise do nothing
        if dynamic_sla:
            record["sla_class"] = dynamic_sla
            logging.debug(
                f'match found applying dynamic sla_class="{dynamic_sla}" for object="{record.get("object")}", key="{key_value}"'
            )


def get_sla_timer(record, sla_classes, sla_default_class):
    """
    Calculates and render the sla_timer
    """

    # a JSON object comntaining a summary for sla information for rendering purposes
    sla_message_json = {}

    # check sla_class (if not in the record, use sla_default_class)
    sla_class = record.get("sla_class", None)
    if not sla_class:
        sla_class = sla_default_class
        record["sla_class"] = sla_default_class

    else:
        # get sla_treshold from sla_classes, if the mentioned sla_class is not found in sla_classes, use the default and replace the record
        if sla_class not in sla_classes:
            sla_class = sla_default_class
            record["sla_class"] = sla_default_class

    # add to sla_message_json
    sla_message_json["sla_class"] = sla_class

    # get sla_threshold from sla_classes
    try:
        sla_threshold = int(sla_classes[sla_class]["sla_threshold"])
    except Exception as e:
        sla_threshold = 86400

    # convert to sla_threshold_duration
    record["sla_threshold_duration"] = convert_seconds_to_duration(sla_threshold)

    # add to record
    record["sla_threshold"] = sla_threshold

    # add to sla_message_json
    sla_message_json["object"] = record.get("object")

    # Calculates
    # for sla, we need to use the current object_state from the KVstore, rather than realtime calculate object_state
    object_state = record.get("kvcurrent_object_state", "red")

    # we will use the realtime object_state to manage a different SLA message if we detect that the KVstore object_state is not yet updated
    realtime_object_state = record.get("object_state", "red")

    sla_message_json["object_state"] = object_state

    if object_state == "red":

        try:
            latest_flip_time = float(record.get("latest_flip_time", 0))
        except Exception as e:
            latest_flip_time = 0

        if latest_flip_time > 0:
            sla_timer = int(round(float(int(time.time()) - latest_flip_time), 0))
        else:
            sla_timer = 0

        # add sla_timer
        record["sla_timer"] = sla_timer
        sla_message_json["sla_timer"] = sla_timer

        # calculate sla_timer_duration
        sla_timer_duration = convert_seconds_to_duration(sla_timer)

        # add
        record["sla_timer_duration"] = convert_seconds_to_duration(sla_timer)
        sla_message_json["sla_timer_duration"] = sla_timer_duration

        # SLA breached
        sla_is_breached = 1 if sla_timer > sla_threshold else 0
        record["sla_is_breached"] = sla_is_breached
        sla_message_json["sla_is_breached"] = sla_is_breached

        # add an sla_message
        if sla_is_breached == 1:

            sla_message = f"SLA breached, the entity has been in a red state for more than {convert_seconds_to_duration(sla_timer)} (sla_class: {sla_class}, sla_class_threshold: {convert_seconds_to_duration(sla_threshold)}, sla_timer_sec: {int(round(sla_timer, 0))} sec, sla_threshold_sec: {sla_threshold} sec)"
            record["sla_message"] = sla_message
            sla_message_json["sla_message"] = sla_message

        else:
            sla_message = f"SLA is not breached, the entity has been in a red state for {convert_seconds_to_duration(sla_timer)} (sla_class: {sla_class}, sla_class_threshold: {convert_seconds_to_duration(sla_threshold)}, sla_timer_sec: {int(round(sla_timer, 0))} sec, sla_threshold_sec: {sla_threshold} sec)"
            record["sla_message"] = sla_message
            sla_message_json["sla_message"] = sla_message

    elif object_state == "green" and realtime_object_state == "red":
        record["sla_timer"] = 0
        sla_message_json["sla_timer"] = 0

        record["sla_timer_duration"] = "0 sec"
        sla_message_json["sla_timer_duration"] = "0 sec"

        record["sla_is_breached"] = 0
        sla_message_json["sla_is_breached"] = 0

        record["sla_message"] = (
            "SLA status refresh is pending, the realtime entity state is red and SLA will be reflected in the next minutes once the KVstore status is updated by trackers"
        )
        sla_message_json["sla_message"] = (
            "SLA status refresh is pending, the realtime entity state is red and SLA will be reflected in the next minutes once the KVstore status is updated by trackers"
        )

    else:
        record["sla_timer"] = 0
        sla_message_json["sla_timer"] = 0

        record["sla_timer_duration"] = "0 sec"
        sla_message_json["sla_timer_duration"] = "0 sec"

        record["sla_is_breached"] = 0
        sla_message_json["sla_is_breached"] = 0

        record["sla_message"] = "SLA is not breached, the entity is not in a red state"
        sla_message_json["sla_message"] = (
            "SLA is not breached, the entity is not in a red state"
        )

    # add sla_message_json to record
    record["sla_message_json"] = sla_message_json
